(*
 * This file is part of Cafesterol.
 * Copyright (C) 2007-2009 Xavier Clerc.
 * Original file (driver/optmain.ml in the Objective Caml source
 * distribution) is Copyright (C) INRIA.
 *
 * Cafesterol is free software; you can redistribute it and/or modify
 * it under the terms of the Q Public License as published by
 * Trolltech (with a change to choice of law).
 *
 * Cafesterol is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Q Public License for more details.
 *
 * You should have received a copy of the Q Public License
 * along with this program.  If not, see
 * <http://trolltech.com/products/qt/licenses/licensing/qpl>.
 *)

open Config
open Clflags
open Jclflags


let output_prefix name =
  let oname =
    match !output_name with
    | None -> name
    | Some n -> if !compile_only then (output_name := None; n) else name in
  Misc.chop_extension_if_any oname

let process_interface_file ppf name =
  Javacompile.interface ppf name (output_prefix name)

let process_implementation_file ppf name =
  let opref = output_prefix name in
  Javacompile.implementation ppf name opref;
  objfiles := (opref ^ ".cmj") :: !objfiles

let process_file ppf name =
  if Filename.check_suffix name ".ml"
  || Filename.check_suffix name ".mlt" then begin
      process_implementation_file ppf name
  end
  else if Filename.check_suffix name !Config.interface_suffix then begin
    let opref = output_prefix name in
    Javacompile.interface ppf name opref;
    if !make_package then objfiles := (opref ^ ".cmi") :: !objfiles
  end
  else if Filename.check_suffix name ".cmj"
       || Filename.check_suffix name ".cmja" then
    objfiles := name :: !objfiles
  else if Filename.check_suffix name ".cmi" && !make_package then
    objfiles := name :: !objfiles
  else if Filename.check_suffix name ".class"
       || Filename.check_suffix name ".jar" then
    ccobjs := name :: !ccobjs
  else if Filename.check_suffix name ".java" then begin
    Javacompile.java_file name;
    ccobjs := (Filename.chop_suffix (Filename.basename name) ".java" ^ ".class")
              :: !ccobjs
  end
  else
    raise(Arg.Bad("don't know what to do with " ^ name))

let print_version_and_library () =
  print_string "The Objective Caml Java-bytecode compiler, version ";
  print_string Config.version; print_newline();
  print_string "Standard library directory: ";
  print_string Config.standard_library; print_newline();
  exit 0

let print_version_string () =
  print_string Config.version; print_newline(); exit 0

let print_standard_library () =
  print_string Config.standard_library; print_newline(); exit 0

let fatal err =
  prerr_endline err;
  exit 2

let extract_output = function
  | Some s -> s
  | None ->
      fatal "Please specify the name of the output file, using option -o"

let default_output = function
  | Some s -> s
  | None -> Jconfig.default_executable_name

let usage = "Usage: ocamljava <options> <files>\nOptions are:"

let show_config () =
  Jconfig.print_config stdout;
  exit 0

let main () =
  native_code := false;
  let ppf = Format.err_formatter in
  try
    Arg.parse ([
       "-a", Arg.Set make_archive, " Build a library";
       "-additional-class", Arg.String (fun s -> additional_classes := s :: !additional_classes),
             " Add a class to the created jar file";
       "-additional-file", Arg.String (fun s -> additional_files := s :: !additional_files),
             " Add a file to the created jar file";
       "-additional-jar", Arg.String (fun s -> additional_jars := s :: !additional_jars),
             " Add a jar file to dependency list";
       "-applet", Arg.Set applet, " Build an applet";
       "-annot", Arg.Set annotations, " Save information in <filename>.annot";
       "-c", Arg.Set compile_only, " Compile only (do not link)";
       "-cadmium-parameter", Arg.String (fun s -> cadmium_parameters := s :: !cadmium_parameters),
         "<param=value>  Add param=value binding to the Cadmium runtime parameters";
       "-classpath", Arg.String (fun s -> class_path := s :: !class_path),
         "<path>  Add <path> to the classpath";
       "-compact", Arg.Clear optimize_for_speed,
             " Optimize code size rather than speed";
       "-config", Arg.Unit show_config,
             " print configuration values and exit";
       "-dtypes", Arg.Set annotations,
             " (deprecated) same as -annot";
       "-for-pack", Arg.String (fun s -> for_package := Some s),
             "<ident>  Generate code that can later be `packed' with\n\
         \     ocamljava -pack -o <ident>.cmj";
       "-g", Arg.Set debug, " Record debugging information for exception backtrace";
       "-i", Arg.Unit (fun () -> print_types := true; compile_only := true),
             " Print inferred interface";
       "-I", Arg.String(fun dir -> include_dirs := dir :: !include_dirs),
             "<dir>  Add <dir> to the list of include directories";
       "-impl", Arg.String (process_implementation_file ppf),
             "<file>  Compile <file> as a .ml file";
       "-inline", Arg.Int(fun n -> inline_threshold := n * 8),
             "<n>  Set aggressiveness of inlining to <n>";
       "-intf", Arg.String (process_interface_file ppf),
             "<file>  Compile <file> as a .mli file";
       "-intf-suffix", Arg.String (fun s -> Config.interface_suffix := s),
             "<file>  Suffix for interface files (default: .mli)";
       "-intf_suffix", Arg.String (fun s -> Config.interface_suffix := s),
             "<file>  (deprecated) same as -intf-suffix";
       "-java-package", Arg.Set_string java_package,
             "<pack>  Set the Java package for produced classes";
       "-javac", Arg.String(fun s -> c_compiler := Some s),
             "<comp>  Use <comp> as the Java compiler";
       "-jopt", Arg.String(fun s -> ccopts := s :: !ccopts),
             "<opt>  Pass option <opt> to the Java compiler";
       "-labels", Arg.Clear classic, " Use commuting label mode";
       "-noassert", Arg.Set noassert, " Don't compile assertion checks";
       "-nobuiltin", Arg.Set no_builtin, " Don't use builtin primitive list";
       "-nolabels", Arg.Set classic, " Ignore non-optional labels in types";
       "-nomerge", Arg.Set no_merge, " Don't merge service descriptors";
       "-nostackmap", Arg.Set no_stackmap, " Don't generate stack maps";
       "-nostdlib", Arg.Set no_std_include,
           " do not add standard directory to the list of include directories";
       "-o", Arg.String(fun s -> output_name := Some s),
             "<file>  Set output file name to <file>";
       "-pack", Arg.Set make_package,
              " Package the given .cmj files into one .cmj";
       "-pp", Arg.String(fun s -> preprocessor := Some s),
             "<command>  Pipe sources through preprocessor <command>";
       "-provider", Arg.String (fun s -> additional_providers := s :: !additional_providers),
             "<classname>  Register <classname> as a primitive provider";
       "-principal", Arg.Set principal,
             " Check principality of type inference";
       "-rectypes", Arg.Set recursive_types,
             " Allow arbitrary recursive types";
       "-scripting", Arg.Set scripting,
             " Compile for Java scripting (for internal use)";
       "-servlet", Arg.String (fun s -> servlet := Some s),
             "<file>  Link as a servlet (passed file is web.xml)";
       "-shared", Arg.Unit (fun () -> shared := true; dlcode := true), 
             " Produce a dynlinkable plugin";
       "-standalone", Arg.Set standalone,
             " Link in standalone mode (no dependency)";
       "-thread", Arg.Set use_threads,
             " Generate code that supports the system threads library";
       "-unsafe", Arg.Set fast,
             " No bounds checking on array and string access";
       "-v", Arg.Unit print_version_and_library,
             " Print compiler version and standard library location and exit";
       "-version", Arg.Unit print_version_string,
             " Print compiler version and exit";
       "-verbose", Arg.Set verbose, " Print calls to external commands";
       "-w", Arg.String (Warnings.parse_options false),
             "<flags>  Enable or disable warnings according to <flags>:\n\
         \032    C/c enable/disable suspicious comment\n\
         \032    D/d enable/disable deprecated features\n\
         \032    E/e enable/disable fragile match\n\
         \032    F/f enable/disable partially applied function\n\
         \032    L/l enable/disable labels omitted in application\n\
         \032    M/m enable/disable overriden methods\n\
         \032    P/p enable/disable partial match\n\
         \032    S/s enable/disable non-unit statement\n\
         \032    U/u enable/disable unused match case\n\
         \032    V/v enable/disable overriden instance variables\n\
         \032    Y/y enable/disable suspicious unused variables\n\
         \032    Z/z enable/disable all other unused variables\n\
         \032    X/x enable/disable all other warnings\n\
         \032    A/a enable/disable all warnings\n\
         \032    default setting is \"Aelz\"";
       "-warn-error" , Arg.String (Warnings.parse_options true),
        "<flags>  Treat the warnings of <flags> as errors, if they are\n\
         \     enabled.  See option -w for the list of flags.\n\
         \     Default setting is \"a\" (warnings are not errors)";
       "-where", Arg.Unit print_standard_library,
         " Print location of standard library and exit";

       "-nopervasives", Arg.Set nopervasives, " (undocumented)";
       "-dparsetree", Arg.Set dump_parsetree, " (undocumented)";
       "-drawlambda", Arg.Set dump_rawlambda, " (undocumented)";
       "-dlambda", Arg.Set dump_lambda, " (undocumented)";
       "-djlambda", Arg.Set dump_jlambda, " (undocumented)";
       "-dminstr", Arg.Set dump_minstr, " (undocumented)";
       "-print-primitives", Arg.Set print_primitives, " (undocumented)";
       "-", Arg.String (process_file ppf),
            "<file>  Treat <file> as a file name (even if it starts with `-')"
      ]) (process_file ppf) usage;
    if !print_primitives then Cadmiumprimitives.init ();
    if List.length (List.filter
                      (fun x -> !x)
                      [make_archive;make_package;shared;compile_only]) > 1
        then fatal "Please specify at most one of -pack, -a, -shared, -c";
    if !make_archive then begin
      Javacompile.init_path();
      let target = extract_output !output_name in
      Javalibrarian.create_archive (List.rev !objfiles) target
    end
    else if !make_package then begin
      Javacompile.init_path();
      let target = extract_output !output_name in
      Javapackager.package_files ppf (List.rev !objfiles) target
    end
    else if !shared then begin
      Javacompile.init_path();
      let target = extract_output !output_name in
      Javalink.link_shared (List.rev !objfiles) target
    end
    else if not !compile_only && !objfiles <> [] then begin
      let target = default_output !output_name in
      Javacompile.init_path();
      Javalink.link (List.rev !objfiles) target
    end;
    exit 0
  with x ->
    Javaerrors.report_error ppf x;
    exit 2

let _ = main ()
