(*
 * This file is part of Cafesterol.
 * Copyright (C) 2007-2009 Xavier Clerc.
 *
 * Cafesterol is free software; you can redistribute it and/or modify
 * it under the terms of the Q Public License as published by
 * Trolltech (with a change to choice of law).
 *
 * Cafesterol is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Q Public License for more details.
 *
 * You should have received a copy of the Q Public License
 * along with this program.  If not, see
 * <http://trolltech.com/products/qt/licenses/licensing/qpl>.
 *)

open Macroinstr
open Bytecodeutils
open Instruct
open Lambda
open Asttypes
open BaristaLibrary
open Utils


type error =
  | Method_too_long of string

exception Error of error


(* compilation state *)

let current_file = ref ""

let current_class = ref ""

let current_function = ref ""

let exception_table = ref []

let garbage_index = ref 0

let current_catches = Hashtbl.create 17

let line_number_table = ref []

let prev_ofs = ref 0

let initial_state = ref StackState.dummy


(* constants *)

let constants = ref []

let reset_constants () =
  constants := []

let add_constant c =
  let res = List.length !constants in
  constants := c :: !constants;
  res

let compile_constants () =
  let rec conv = function
    | Const_base (Const_int i) -> Obj.repr i
    | Const_base (Const_char c) -> Obj.repr c
    | Const_base (Const_string s) -> Obj.repr s
    | Const_base (Const_float s) -> Obj.repr (float_of_string s)
    | Const_base (Const_int32 i) -> Obj.repr i
    | Const_base (Const_int64 i) -> Obj.repr i
    | Const_base (Const_nativeint i) -> Obj.repr i
    | Const_pointer i -> Obj.repr i
    | Const_immstring s -> Obj.repr s
    | Const_block (tag, fields) ->
        let block = Obj.new_block tag (List.length fields) in
        let pos = ref 0 in
        List.iter
          (fun c ->
            Obj.set_field block !pos (conv c);
            incr pos)
          fields;
        block
    | Const_float_array fields ->
        Obj.repr (Array.of_list (List.map float_of_string fields)) in
  let towrite = List.rev_map conv !constants in
  let array = Array.make (List.length towrite) (Obj.magic 0) in
  let i = ref 0 in
  List.iter
    (fun c ->
      array.(!i) <- c;
      incr i)
    towrite;
  Marshal.to_string array []

let compile_constant boxed_result c =
  let make_long x =
    match boxed_result with
    | Boxed -> push_boxed_int x
    | Unboxed -> push_int x in
  match c with
  | Const_base (Const_int i)
  | Const_pointer i ->
      make_long i
  | Const_base (Const_char c) ->
      make_long (Char.code c)
  | _ ->
      let idx = add_constant c in
      let idx_index, idx_index_sz, idx_index_st = push_int idx in
      [ Instruction.ALOAD_0 ]
      @ [ Instruction.LDC_W (`Class_or_interface (make_class !current_class)) ]
      @ idx_index
      @ meth_getConstant
      @ (if boxed_result = Unboxed then meth_asLong else []),
      1 + 3 + idx_index_sz + meth_size
        + (if boxed_result = Unboxed then meth_size else 0),
      2 + idx_index_st


(* expression compilation *)

let parameters n =
  let rec params acc = function
    | 0 -> acc
    | n -> params ((`Class class_Value) :: acc) (pred n) in
  params [] n

let check_jump ofs =
  if ofs >= -32768 && ofs <= 32767 then
    ofs
  else
    raise (Error (Method_too_long !current_function))

let add_debug_info ofs dbg =
  if dbg != Debuginfo.none && dbg.Debuginfo.dinfo_file = !current_file then
    begin
      line_number_table :=
        (!prev_ofs, dbg.Debuginfo.dinfo_line) :: !line_number_table;
      prev_ofs := ofs
    end

let take n l =
  let rec take' n acc l =
    match n, l with
    | 0, _ -> acc
    | _, hd :: tl -> take' (pred n) (hd :: acc) tl
    | _ -> assert false in
  List.rev (take' n [] l)

let with_state state (x, y, z) = (x, y, z, [], state)

let rec compile_expression state stack is_tail boxed_result ofs = function
  | Mconst c ->
      with_state
        (StackState.push_kind boxed_result state)
        (compile_constant boxed_result c)
  | Moffset (expr, n) ->
      let expr_instrs, expr_sz, expr_st, expr_sl, _ =
        compile_expression state stack false Boxed ofs expr in
      let int_instrs, int_sz, int_st = push_int n in
      expr_instrs
      @ meth_asBlock
      @ int_instrs
      @ meth_offset
      @ (if boxed_result = Unboxed then meth_asLong else []),
      expr_sz + meth_size + int_sz + meth_size
        + (if boxed_result = Unboxed then meth_size else 0),
      max expr_st (1 + int_st),
      expr_sl,
      (StackState.push_kind boxed_result state)
  | Mdynamicoffset (expr, n) ->
      let expr_instrs, expr_sz, expr_st, expr_sl, _ =
        compile_expression state stack false Boxed ofs expr in
      let int_instrs, int_sz, int_st, int_sl, _ =
        compile_expression (StackState.push_block state) (Boxed :: stack) false Unboxed (ofs + expr_sz + meth_size + 1 + meth_size + meth_size) n in
      expr_instrs
      @ meth_asBlock
      @ [ Instruction.ICONST_0 ]
      @ meth_get
      @ meth_asBlock
      @ int_instrs
      @ meth_get
      @ (if boxed_result = Unboxed then meth_asLong else []),
      expr_sz + meth_size + 1 + meth_size + meth_size + int_sz + meth_size
        + (if boxed_result = Unboxed then meth_size else 0),
      max 2 (max expr_st (1 + int_st)),
      expr_sl
      @ (StackState.translate (expr_sz + meth_size + 1 + meth_size + meth_size) int_sl),
      (StackState.push_kind boxed_result state)
  | Mcreateglobal (id, sz) ->
      let size_instrs, size_sz, size_st = push_int sz in
      [ Instruction.ALOAD_0 ;
        Instruction.LDC_W (`String (utf8_of_string id)) ]
      @ size_instrs
      @ meth_createGlobal,
      1 + 3 + size_sz + meth_size,
      2 + size_st,
      [],
      state
  | Mclosure (2, ((cls, func), 1) :: [], []) ->
      [ Instruction.ALOAD_0 ;
        Instruction.LDC_W (`Class_or_interface (make_class cls)) ;
        Instruction.LDC_W (`String (normalize_function_name func)) ]
      @ meth_registerClosure1,
      1 + 3 + 3 + meth_size,
      3,
      [],
      (StackState.push_value state)
  | Mclosure (3, ((cls, func), arity) :: [], []) ->
      let arity_instrs, arity_sz, arity_st = push_int arity in
      [ Instruction.ALOAD_0 ;
        Instruction.LDC_W (`Class_or_interface (make_class cls)) ;
        Instruction.LDC_W (`String (normalize_function_name func)) ]
      @ arity_instrs
      @ meth_registerClosureN,
      1 + 3 + 3 + arity_sz + meth_size,
      3 + arity_st,
      [],
      (StackState.push_value state)
  | Mclosure (sz, ((cls, func), arity) :: others, vars) ->
      let size_instrs, size_sz, size_st = push_int sz in
      let load_instrs, load_sz, _ = aload !garbage_index in
      let store_instrs, store_sz, _ = astore !garbage_index in
      let state' = StackState.store_block !garbage_index state in
      let first_instrs, first_sz, first_st, next =
        compile_first_closure cls func arity in
      let others_instrs, others_sz, others_st, index =
        compile_other_closures next others in
      let vars_instrs, vars_sz, vars_st, vars_sl, _ =
        compile_expression_list (StackState.push_block state') (Boxed :: stack) Boxed (ofs + size_sz + meth_size + 1 + 1 + meth_size + store_sz + first_sz + others_sz + load_sz) vars in
      let index_instrs, index_sz, index_st = push_int index in
      let n_instrs, n_sz, n_st = push_int (List.length vars) in
      let fold_instrs = ref [] in
      let fold_sz = ref 0 in
      let remaining = ref (List.length vars) in
      while !remaining > 0 do
        let n = min 4 !remaining in
        fold_instrs := !fold_instrs
          @ [ Instruction.INVOKESTATIC
                (class_AbstractNativeRunner,
                 make_method ("foldClosure" ^ (string_of_int n)),
                 ((parameters n) @ [`Class class_Closure],
                  `Class class_Closure)) ];
        fold_sz := !fold_sz + meth_size;
        remaining := !remaining - n
      done;
      size_instrs
      @ meth_createClosure
      @ [ Instruction.DUP ;
          Instruction.DUP ]
      @ meth_setCustom
      @ store_instrs
      @ first_instrs
      @ others_instrs
      @ load_instrs
      @ vars_instrs
      @ index_instrs
      @ n_instrs
      @ meth_createClosureVars
      @ !fold_instrs
      @ meth_returnClosure,
      size_sz + meth_size + 1 + 1 + meth_size + store_sz + first_sz + others_sz
        + load_sz + vars_sz + index_sz + n_sz + meth_size + !fold_sz + meth_size,
      List.fold_left
        max
        0
        [ size_st ;
          3 ;
          first_st ;
          others_st ;
          1 + vars_st ;
          1 + (List.length vars) + index_st + n_st],
      (StackState.translate (size_sz + meth_size + 1 + 1 + meth_size + store_sz + first_sz + others_sz + load_sz) vars_sl),
      (StackState.push_value state')
  | Mclosure _ ->
      Misc.fatal_error "Bytecodegen.compile_expression: invalid closure"
  | Mreadlocal idx ->
      if boxed_result = Boxed then
        with_state
          (StackState.push_value state)
          (aload idx)
      else
        let aload_instrs, aload_sz, aload_st = aload idx in
        aload_instrs
        @ meth_asLong,
        aload_sz + meth_size,
        aload_st,
        [],
        (StackState.push_int state)
  | Mreadlocalint idx ->
      if boxed_result = Unboxed then
        with_state
          (StackState.push_int state)
          (iload idx)
      else
        let iload_instrs, iload_sz, iload_st = iload idx in
        iload_instrs
        @ meth_createFromLong,
        iload_sz + meth_size,
        iload_st,
        [],
        (StackState.push_value state)
  | Mwritelocal (idx, expr) ->
      let expr_instrs, expr_sz, expr_st, expr_sl, _ =
        compile_expression state stack false Boxed ofs expr in
      let astore_instrs, astore_sz, astore_st = astore idx in
      expr_instrs
      @ astore_instrs,
      expr_sz + astore_sz,
      max expr_st (1 + astore_st),
      expr_sl,
      (StackState.store_value idx state)
  | Mwritelocalint (idx, expr) ->
      let expr_instrs, expr_sz, expr_st, expr_sl, _ =
        compile_expression state stack false Unboxed ofs expr in
      let istore_instrs, istore_sz, istore_st = istore idx in
      expr_instrs
      @ istore_instrs,
      expr_sz + istore_sz,
      max expr_st (1 + istore_st),
      expr_sl,
      (StackState.store_int idx state)
  | Mcall (call, (cls, func), args, dbg) ->
      add_debug_info ofs dbg;
      if is_tail && cls = !current_class && func = !current_function then
        let args_instrs, args_sz, args_st, args_sl, _ =
          compile_expression_list_to_locals (StackState.push_null state) (Boxed :: stack) (succ ofs) args in
        [ Instruction.ACONST_NULL ]
        @ args_instrs
        @ [ Instruction.POP ]
        @ [ Instruction.GOTO (check_jump ~-(ofs + 1 + args_sz + 1)) ]
        @ field_UNIT
        @ [ Instruction.ARETURN ],
        1 + args_sz + 1 + 3 + field_size + 1,
        1 + args_st,
        [(-ofs), !initial_state]
        @ (StackState.translate 1 args_sl)
        @ [(1 + args_sz + 1 + 3), (StackState.empty state)],
        (StackState.push_kind boxed_result (StackState.empty state))
      else begin
        let arity = List.length args in
        let args_instrs, args_sz, args_st, args_sl, _ =
          compile_expression_list (StackState.push_runner state) (Boxed :: stack) Boxed (succ ofs) args in
        [ Instruction.ALOAD_0 ]
        @ args_instrs
        @ [ Instruction.INVOKESTATIC
              (make_class cls,
               make_method func,
               ((`Class (if call = Native then class_AbstractNativeRunner else class_CodeRunner))
                :: (parameters arity), `Class class_Value)) ]
        @ (if boxed_result = Unboxed then meth_asLong else []),
        1 + args_sz + meth_size
          + (if boxed_result = Unboxed then meth_size else 0),
        1 + args_st,
        (StackState.translate 1 args_sl),
        (StackState.push_kind boxed_result state)
      end
  | Mprim (prim, args, dbg) ->
      add_debug_info ofs dbg;
      (try
        compile_expression state stack false boxed_result ofs (Bytecodeprimitive.as_mcall prim args)
      with
      | Bytecodeprimitive.Not_a_call ->
          match prim, args with
          | Pidentity, [arg] ->
              compile_expression state stack false boxed_result ofs arg
          | Pignore, [arg] ->
              let arg_instrs, arg_sz, arg_st, arg_sl, _ =
                compile_expression state stack false Boxed ofs arg in
              arg_instrs
              @ [ Instruction.POP ]
              @ field_UNIT,
              arg_sz + 1 + field_size,
              arg_st,
              arg_sl,
              (StackState.push_value (StackState.pop (StackState.push_value state)))
          | Pnot, [arg] ->
              (match boxed_result with
              | Boxed ->
                  let arg_instrs, arg_sz, arg_st, arg_sl, _ =
                    compile_expression state stack false Boxed ofs arg in
                  arg_instrs
                  @ field_FALSE
                  @ [ Instruction.IF_ACMPEQ (3 + field_size + 3) ]
                  @ field_FALSE
                  @ [ Instruction.GOTO (3 + field_size) ]
                  @ field_TRUE,
                  arg_sz + field_size + 3 + field_size + 3 + field_size,
                  arg_st + 1,
                  arg_sl
                  @ [(arg_sz + field_size + 3 + field_size + 3), state]
                  @ [(arg_sz + field_size + 3 + field_size + 3 + field_size), (StackState.push_value state)],
                  (StackState.push_value state)
              | Unboxed ->
                  let arg_instrs, arg_sz, arg_st, arg_sl, _ =
                    compile_expression state stack false Unboxed ofs arg in
                  arg_instrs
                  @ [ Instruction.IFEQ (3 + 1 + 3) ;
                      Instruction.ICONST_0 ;
                      Instruction.GOTO (3 + 1) ;
                      Instruction.ICONST_1 ],
                  arg_sz + 3 + 1 + 3 + 1,
                  arg_st,
                  arg_sl
                  @ [(arg_sz + 3 + 1 + 3), state]
                  @ [(arg_sz + 3 + 1 + 3 + 1), (StackState.push_int state)],
                  (StackState.push_int state))
          | Psequand, [expr1; expr2] ->
              let expr1_instrs, expr1_sz, expr1_st, expr1_sl, _ =
                compile_expression state stack false Unboxed ofs expr1 in
              let expr2_instrs, expr2_sz, expr2_st, expr2_sl, _ =
                compile_expression state stack false Unboxed (ofs + expr1_sz + 3) expr2 in
              expr1_instrs
              @ [ Instruction.IFEQ (check_jump (3 + expr2_sz + 3)) ]
              @ expr2_instrs
              @ [ Instruction.GOTO (3 + 1) ;
                  Instruction.ICONST_0 ]
              @ (if boxed_result = Boxed then meth_createFromLong else []),
              expr1_sz + 3 + expr2_sz + 3 + 1
                + (if boxed_result = Boxed then meth_size else 0),
              max expr1_st expr2_st,
              expr1_sl
              @ (StackState.translate (expr1_sz + 3) expr2_sl)
              @ [(expr1_sz + 3 + expr2_sz + 3, state);
                 (expr1_sz + 3 + expr2_sz + 3 + 1, (StackState.push_int state))],
              (StackState.push_kind boxed_result state)
          | Psequor, [expr1; expr2] ->
              let expr1_instrs, expr1_sz, expr1_st, expr1_sl, _ =
                compile_expression state stack false Unboxed ofs expr1 in
              let expr2_instrs, expr2_sz, expr2_st, expr2_sl, _ =
                compile_expression state stack false Unboxed (ofs + expr1_sz + 3) expr2 in
              expr1_instrs
              @ [ Instruction.IFNE (check_jump (3 + expr2_sz + 3)) ]
              @ expr2_instrs
              @ [ Instruction.GOTO (3 + 1) ;
                  Instruction.ICONST_1 ]
              @ (if boxed_result = Boxed then meth_createFromLong else []),
              expr1_sz + 3 + expr2_sz + 3 + 1
                + (if boxed_result = Boxed then meth_size else 0),
              max expr1_st expr2_st,
              expr1_sl
              @ (StackState.translate (expr1_sz + 3) expr2_sl)
              @ [(expr1_sz + 3 + expr2_sz + 3, state);
                 (expr1_sz + 3 + expr2_sz + 3 + 1, (StackState.push_int state))],
              (StackState.push_kind boxed_result state)
          | Pstringrefu, [str; idx] ->
              let str_instrs, str_sz, str_st, str_sl, _ =
                compile_expression state stack false Boxed ofs str in
              let idx_instrs, idx_sz, idx_st, idx_sl, _ =
                compile_expression (StackState.push_block state) (Boxed :: stack) false Unboxed (ofs + str_sz + meth_size) idx in
              str_instrs
              @ meth_asBlock
              @ idx_instrs
              @ meth_getUnsignedByte
              @ (if boxed_result = Boxed then meth_createFromLong else []),
              str_sz + meth_size + idx_sz + meth_size
                + (if boxed_result = Boxed then meth_size else 0),
              max str_st (1 + idx_st),
              str_sl
              @ (StackState.translate (str_sz + meth_size) idx_sl),
              (StackState.push_kind boxed_result state)
          | Pstringsetu, [str; idx; v] ->
              let str_instrs, str_sz, str_st, str_sl, _ =
                compile_expression state stack false Boxed ofs str in
              let idx_instrs, idx_sz, idx_st, idx_sl, _ =
                compile_expression (StackState.push_block state) (Boxed :: stack) false Unboxed (ofs + str_sz + meth_size) idx in
              let v_instrs, v_sz, v_st, v_sl, _ =
                compile_expression (StackState.push_int (StackState.push_block state)) (Unboxed :: Boxed :: stack) false Unboxed (ofs + str_sz + meth_size + idx_sz) v in
              str_instrs
              @ meth_asBlock
              @ idx_instrs
              @ v_instrs
              @ [ Instruction.SIPUSH 0x00FF ;
                  Instruction.IAND ]
              @ meth_setUnsignedByte
              @ field_UNIT,
              str_sz + meth_size + idx_sz + v_sz + 3 + 1 + meth_size + field_size,
              max (max str_st (1 + idx_st)) (max (2 + v_st) 4),
              str_sl
              @ (StackState.translate (str_sz + meth_size) idx_sl)
              @ (StackState.translate (str_sz + meth_size + idx_sz) v_sl),
              (StackState.push_value state)
          | (Pmakearray _), [] ->
              with_state
                (StackState.push_value state)
                (compile_constant Boxed (Const_block (0, [])))
          | (Pmakearray kind), args ->
              (match kind with
              | Pintarray | Paddrarray ->
                  compile_block state stack 0 Mutable ofs args
              | Pfloatarray ->
                  let size = List.length args in
                  let args_instrs, args_sz, args_st, args_sl, _ =
                    compile_expression_list state stack Boxed ofs args in
                  let size_instrs, size_sz, size_st = push_int size in
                  let fold_instrs = ref [] in
                  let fold_sz = ref 0 in
                  let remaining = ref size in
                  while !remaining > 0 do
                    let n = min 4 !remaining in
                    fold_instrs := !fold_instrs
                      @ [ Instruction.INVOKESTATIC
                            (class_AbstractNativeRunner,
                             make_method ("foldValues" ^ (string_of_int n)),
                             ((parameters n) @ [`Class class_Values],
                              `Class class_Values)) ];
                    fold_sz := !fold_sz + meth_size;
                    remaining := !remaining - n
                  done;
                  args_instrs
                  @ size_instrs
                  @ meth_createDoubles
                  @ !fold_instrs
                  @ meth_returnDoubles,
                  args_sz + size_sz + meth_size + !fold_sz + meth_size,
                  max args_st (size + size_st),
                  args_sl,
                  (StackState.push_value state)
              | Pgenarray ->
                  let block_instrs, block_sz, block_st, block_sl, _ =
                    compile_block state stack 0 Mutable ofs args in
                  block_instrs
                  @ [ Instruction.ALOAD_0 ;
                      Instruction.SWAP ]
                  @ meth_caml_make_array,
                  block_sz + 1 + 1 + meth_size,
                  1 + block_st,
                  block_sl,
                  (StackState.push_value state))
          | (Parrayrefu _), [arr; idx] ->
              let arr_instrs, arr_sz, arr_st, arr_sl, _ =
                compile_expression state stack false Boxed ofs arr in
              let idx_instrs, idx_sz, idx_st, idx_sl, _ =
                compile_expression (StackState.push_value state) (Boxed :: stack) false Unboxed (ofs + arr_sz + meth_size) idx in
              arr_instrs
              @ meth_asBlock
              @ idx_instrs
              @ meth_get
              @ (if boxed_result = Unboxed then meth_asLong else []),
              arr_sz + meth_size + idx_sz + meth_size + (if boxed_result = Unboxed then meth_size else 0),
              max arr_st (1 + idx_st),
              arr_sl
              @ (StackState.translate (arr_sz + meth_size) idx_sl),
              (StackState.push_kind boxed_result state)
          | (Parraysetu _), [arr; idx; v] ->
              let arr_instrs, arr_sz, arr_st, arr_sl, _ =
                compile_expression state stack false Boxed ofs arr in
              let idx_instrs, idx_sz, idx_st, idx_sl, _ =
                compile_expression (StackState.push_value state) (Boxed :: stack) false Unboxed (ofs + arr_sz + meth_size) idx in
              let v_instrs, v_sz, v_st, v_sl, _ =
                compile_expression (StackState.push_int (StackState.push_value state)) (Unboxed :: Boxed :: stack) false Boxed (ofs + arr_sz + meth_size + idx_sz) v in
              arr_instrs
              @ meth_asBlock
              @ idx_instrs
              @ v_instrs
              @ meth_set
              @ field_UNIT,
              arr_sz + meth_size + idx_sz + v_sz + meth_size + field_size,
              max arr_st (max (1 + idx_st) (2 + v_st)),
              arr_sl
              @ (StackState.translate (arr_sz + meth_size) idx_sl)
              @ (StackState.translate (arr_sz + meth_size + idx_sz) v_sl),
              (StackState.push_value state)
          | (Pmakeblock (tag, mut)), args ->
              compile_block state stack tag mut ofs args
          | Pisint, [arg] ->
              let arg_instrs, arg_sz, arg_st, arg_sl, _ =
                compile_expression state stack false Boxed ofs arg in
              (match boxed_result with
              | Boxed ->
                  arg_instrs
                  @ meth_isLong
                  @ meth_createFromLong,
                  arg_sz + meth_size + meth_size,
                  arg_st,
                  arg_sl,
                  (StackState.push_value state)
              | Unboxed ->
                  arg_instrs
                  @ meth_isLong,
                  arg_sz + meth_size,
                  arg_st,
                  arg_sl,
                  (StackState.push_int state))
          | (Pintcomp cmp), [arg1; arg2] ->
              let arg1_instrs, arg1_sz, arg1_st, arg1_sl, _ =
                compile_expression state stack false Boxed ofs arg1 in
              let arg2_instrs, arg2_sz, arg2_st, arg2_sl, _ =
                compile_expression (StackState.push_value state) (Boxed :: stack) false Boxed (ofs + arg1_sz) arg2 in
              arg1_instrs
              @ arg2_instrs
              @ (match cmp with
              | Ceq -> meth_equalValues
              | Cneq -> meth_notEqualValues
              | Clt -> meth_lowerThanValue
              | Cgt -> meth_greaterThanValue
              | Cle -> meth_lowerEqualValue
              | Cge -> meth_greaterEqualValue)
              @ (if boxed_result = Boxed then meth_createFromLong else []),
              arg1_sz + arg2_sz + meth_size
                + (if boxed_result = Boxed then meth_size else 0),
              max arg1_st (arg2_st + 1),
              arg1_sl
              @ (StackState.translate arg1_sz arg2_sl),
              (StackState.push_kind boxed_result state)
          | Pisout, [arg1; arg2] ->
              let arg1_instrs, arg1_sz, arg1_st, arg1_sl, _ =
                compile_expression state stack false Boxed ofs arg1 in
              let arg2_instrs, arg2_sz, arg2_st, arg2_sl, _ =
                compile_expression (StackState.push_value state) (Boxed :: stack) false Boxed (ofs + arg1_sz) arg2 in
              arg1_instrs
              @ arg2_instrs
              @ meth_isOutValue
              @ (if boxed_result = Boxed then meth_createFromLong else []),
              arg1_sz + arg2_sz + meth_size
                + (if boxed_result = Boxed then meth_size else 0),
              max arg1_st (arg2_st + 1),
              arg1_sl
              @ (StackState.translate arg1_sz arg2_sl),
              (StackState.push_kind boxed_result state)
          | _ ->
              let boxed = Bytecodeprimitive.parameters_type prim in
              let args_instrs, args_sz, args_st, args_sl, args_s =
                compile_expression_list state stack boxed ofs args in
              let prim_instrs, prim_sz, prim_st, prim_sl =
                Bytecodeprimitive.compile args_s prim in
              let conv_instrs, conv_sz = match boxed, boxed_result with
              | Boxed, Unboxed ->
                  meth_asLong, meth_size
              | Unboxed, Boxed ->
                  meth_createFromLong, meth_size
              | Boxed, Boxed | Unboxed, Unboxed -> [], 0 in
              args_instrs
              @ prim_instrs
              @ conv_instrs,
              args_sz + prim_sz + conv_sz,
              max args_st ((List.length args) + prim_st),
              args_sl
              @ (StackState.translate args_sz prim_sl),
              (StackState.push_kind boxed_result state))
  | Mapply (args, dbg) ->
      add_debug_info ofs dbg;
      let arity = pred (List.length args) in
      if (arity <= 4) then begin
        let args_instrs, args_sz, args_st, args_sl, _ =
          compile_expression_list (StackState.push_runner state) (Boxed :: stack) Boxed (succ ofs) args  in
        [ Instruction.ALOAD_0 ]
        @ args_instrs
        @ [ Instruction.INVOKEVIRTUAL
              ((`Class_or_interface class_AbstractNativeRunner),
               make_method ("apply" ^ (string_of_int (arity))),
               ((parameters (succ arity)), `Class class_Value)) ]
        @ (if boxed_result = Unboxed then meth_asLong else [])
        @ [ Instruction.ALOAD_0 ]
        @ meth_checkSignals,
        1 + args_sz + meth_size + (if boxed_result = Unboxed then meth_size else 0) + 1 + meth_size,
        max (1 + args_st) 2,
        (StackState.translate 1 args_sl),
        (StackState.push_kind boxed_result state)
      end else begin
        let args_instrs, args_sz, args_st, args_sl, _ =
          compile_expression_list state stack Boxed ofs args  in
        let fold_instrs = ref [] in
        let fold_sz = ref 0 in
        let current_arity = ref arity in
        while !current_arity > 0 do
          let reduce = min !current_arity 4 in
          fold_instrs := !fold_instrs
            @ [ Instruction.INVOKESTATIC
                  (class_AbstractNativeRunner,
                   make_method ("foldApplication" ^ (string_of_int reduce)),
                   ((parameters reduce) @ [`Class class_Application], `Class class_Application)) ];
          current_arity := !current_arity - reduce;
          fold_sz := !fold_sz + meth_size
        done;
        let arity_instrs, arity_sz, arity_st = push_int arity in
        args_instrs
        @ [ Instruction.ALOAD_0 ;
            Instruction.SWAP ]
        @ arity_instrs
        @ meth_createApplication
        @ !fold_instrs
        @ meth_returnApplication
        @ (if boxed_result = Unboxed then meth_asLong else [])
        @ [ Instruction.ALOAD_0 ]
        @ meth_checkSignals,
        args_sz + 2 + arity_sz + meth_size + !fold_sz + meth_size
          + (if boxed_result = Unboxed then meth_size else 0)
          + 1 + meth_size,
        max (args_st + 1 + arity_st) 2,
        args_sl,
        (StackState.push_kind boxed_result state)
      end
  | Msequence (e1, e2) ->
      let e1_instrs, e1_sz, e1_st, e1_sl, e1_s =
        compile_expression state stack false Boxed ofs e1 in
      let stack' = if (StackState.size e1_s) > (StackState.size state) then
        (if boxed_result = Boxed then Boxed :: stack else Unboxed :: stack)
      else
        stack in
      let e2_instrs, e2_sz, e2_st, e2_sl, e2_s =
        compile_expression e1_s stack' is_tail boxed_result (ofs + e1_sz) e2 in
      e1_instrs
      @ e2_instrs,
      e1_sz + e2_sz,
      max e1_st e2_st,
      e1_sl
      @ (StackState.translate e1_sz e2_sl),
      e2_s
  | Mifthenelse (cond, ifso, ifno) ->
      let cond_instrs, cond_sz, cond_st, cond_sl, _ =
        compile_expression state stack false Unboxed ofs cond in
      let ifso_instrs, ifso_sz, ifso_st, ifso_sl, ifso_s =
        compile_expression state stack is_tail boxed_result (ofs + cond_sz + 3) ifso in
      let ifno_instrs, ifno_sz, ifno_st, ifno_sl, ifno_s =
        compile_expression state stack is_tail boxed_result (ofs + cond_sz + 3 + ifso_sz + 3) ifno in
      cond_instrs
      @ [ Instruction.IFEQ (check_jump (3 + ifso_sz + 3)) ]
      @ ifso_instrs
      @ [ Instruction.GOTO (check_jump (3 + ifno_sz)) ]
      @ ifno_instrs,
      cond_sz + 3 + ifso_sz + 3 + ifno_sz,
      max cond_st (max ifso_st ifno_st),
      cond_sl
      @ (StackState.translate (cond_sz + 3) ifso_sl)
      @ [(cond_sz + 3 + ifso_sz), ifso_s]
      @ [(cond_sz + 3 + ifso_sz + 3), state]
      @ (StackState.translate (cond_sz + 3 + ifso_sz + 3) ifno_sl)
      @ [(cond_sz + 3 + ifso_sz + 3 + ifno_sz), ifso_s]
      @ [(cond_sz + 3 + ifso_sz + 3 + ifno_sz), ifno_s],
      (StackState.generalize ifso_s ifno_s)
  | Mwhile (cond, body) ->
      let body_instrs, body_sz, body_st, body_sl, _ =
        compile_expression state stack false Boxed (ofs + 3 + 1 + meth_size) body in
      let cond_instrs, cond_sz, cond_st, cond_sl, _ =
        compile_expression state stack false Unboxed (ofs + 3 + 1 + meth_size + body_sz) cond in
      [ Instruction.GOTO (check_jump (3 + 1 + meth_size + body_sz)) ;
        Instruction.ALOAD_0 ]
      @ meth_checkSignals
      @ body_instrs
      @ cond_instrs
      @ [ Instruction.IFNE (check_jump ~-(1 + meth_size + body_sz + cond_sz)) ],
      3 + 1 + meth_size + body_sz + cond_sz + 3,
      max 2 (max body_st (1 + cond_st)),
      [(3, state)]
      @ (StackState.translate (3 + 1 + meth_size) body_sl)
      @ [(3 + 1 + meth_size + body_sz), state]
      @ (StackState.translate (3 + 1 + meth_size + body_sz) cond_sl),
      state
  | Mfor (idx, dir, bound, body) ->
      let body_instrs, body_sz, body_st, body_sl, _ =
        compile_expression state stack false Boxed (ofs + 3 + 1 + meth_size) body in
      let iinc_instrs, iinc_sz, _ =
        iinc idx (if dir = Upto then 1 else -1) in
      let idx_instrs, idx_sz, _ = iload idx in
      let bound_instrs, bound_sz, _ = iload bound in
      [ Instruction.GOTO (check_jump (3 + 1 + meth_size + body_sz + iinc_sz)) ;
        Instruction.ALOAD_0 ]
      @ meth_checkSignals
      @ body_instrs
      @ iinc_instrs
      @ idx_instrs
      @ bound_instrs
      @ (if dir = Upto then
        [ Instruction.IF_ICMPLE (check_jump ~-(bound_sz + idx_sz + iinc_sz + body_sz + meth_size + 1)) ]
      else
        [ Instruction.IF_ICMPGE (check_jump ~-(bound_sz + idx_sz + iinc_sz + body_sz + meth_size + 1)) ]),
      3 + 1 + meth_size + body_sz + iinc_sz + idx_sz + bound_sz + 3,
      max 2 body_st,
      [(3, state)]
      @ (StackState.translate (3 + 1 + meth_size) body_sl)
      @ [(3 + 1 + meth_size + body_sz + iinc_sz), state],
      state
  | Munit ->
      with_state
        (StackState.push_kind boxed_result state) 
        (match boxed_result with
        | Boxed -> field_UNIT, field_size, field_stack
        | Unboxed -> (push_int 0))
  | Mpoptolocal idx ->
      with_state (StackState.store_value idx (StackState.pop state)) (astore idx)
  | Mpop ->
      [ Instruction.POP ], 1, 0, [], (StackState.pop state)
  | Mnop ->
      [], 0, 0, [], state
  | Mtrywith (body, idx, handler, backup) ->
      let gload_instrs, gload_sz, _ = aload !garbage_index in
      let gstore_instrs, gstore_sz, _ = astore !garbage_index in
      let save_instrs = ref [] in
      let restore_instrs = ref [] in
      let backup_size = ref 0 in
      let state' = ref state in
      assert ((List.length backup) >= (List.length stack));
      List.iter2
        (fun idx kind ->
          let sav, sz, _ =
            match kind with Boxed -> astore idx | Unboxed -> istore idx in
          let rst, _, _ =
            match kind with Boxed -> aload idx | Unboxed -> iload idx in
          save_instrs := sav :: !save_instrs;
          restore_instrs := rst :: !restore_instrs;
          backup_size := sz + !backup_size)
        (take (List.length stack) backup)
        stack;
      List.iter2
        (fun idx kind ->
          match kind with
          | Boxed -> state' := StackState.pop (StackState.store_top idx !state')
          | Unboxed -> state' := StackState.pop (StackState.store_int idx !state'))
        (List.rev (take (List.length stack) backup))
        stack;        
      let body_instrs, body_sz, body_st, body_sl, body_s =
        compile_expression !state' [] false Boxed (ofs + !backup_size) body in
      let store_instrs, store_sz, _ = astore idx in
      let handler_instrs, handler_sz, handler_st, handler_sl, handler_s =
        compile_expression (StackState.store_value idx !state') [] false Boxed (ofs + !backup_size + body_sz + 3 + 1 + meth_size + store_sz) handler in
      exception_table := (ofs + !backup_size,
                          ofs + !backup_size + body_sz,
                          ofs + !backup_size + body_sz + 3)
        :: !exception_table;
      let next_state = ref (StackState.generalize body_s handler_s) in
      List.iter2
        (fun idx _ ->
          next_state := StackState.push_index idx !next_state)
        (List.rev (take (List.length stack) backup))
        (List.rev stack);
      (List.concat !save_instrs)
      @ body_instrs
      @ [ Instruction.GOTO (check_jump (3 + 1 + meth_size + store_sz + handler_sz)) ]
      @ [ Instruction.ALOAD_0 ]
      @ meth_asValue
      @ store_instrs
      @ handler_instrs
      @ (if !backup_size <> 0 then
        gstore_instrs
        @ (List.concat (List.rev !restore_instrs))
        @ gload_instrs
      else
        [])
      @ (if boxed_result = Unboxed then meth_asLong else []),
      !backup_size + body_sz + 3 + 1 + meth_size + store_sz + handler_sz
        + (if !backup_size <> 0 then gstore_sz + !backup_size + gload_sz else 0)
        + (if boxed_result = Unboxed then meth_size else 0),
      1 + (max body_st handler_st),
      (StackState.translate !backup_size body_sl)
      @ [(!backup_size + body_sz + 3, StackState.only_exception !state')]
      @ (StackState.translate (!backup_size + body_sz + 3 + 1 + meth_size + store_sz) handler_sl)
      @ [(!backup_size + body_sz + 3 + 1 + meth_size + store_sz + handler_sz),
         (StackState.generalize body_s handler_s)],
        !next_state
  | Mstaticfail (nfail, args) ->
      let args_instrs, args_sz, args_st, args_sl, args_s =
        compile_expression_list state stack Boxed ofs args in
      let dest = check_jump ((Hashtbl.find current_catches nfail) - (ofs + args_sz)) in
      args_instrs
      @ [ Instruction.GOTO dest ]
      @ field_UNIT
      @ (if boxed_result = Unboxed then meth_asLong else []),
      args_sz + 3 + field_size + (if boxed_result = Unboxed then meth_size else 0),
      max 1 args_st,
      [ 0, state ]
      @ args_sl
      @ [ (dest + args_sz), args_s ]
      @ [ (args_sz + 3), state ],
      (StackState.push_kind boxed_result state)
  | Mstaticcatch (nfail, body, n, handler) ->
      Hashtbl.add current_catches nfail (ofs + 3);
      let handler_instrs, handler_sz, handler_st, handler_sl, handler_s =
        compile_expression (StackState.push_values n state) stack false boxed_result (ofs + 3) handler in
      let body_instrs, body_sz, body_st, body_sl, body_s =
        compile_expression state stack false boxed_result (ofs + 3 + handler_sz + 3) body in
      [ Instruction.GOTO (check_jump (3 + handler_sz + 3)) ]
      @ handler_instrs
      @ [ Instruction.GOTO (check_jump (3 + body_sz)) ]
      @ body_instrs,
      3 + handler_sz + 3 + body_sz,
      max handler_st body_st,
      (StackState.translate 3 handler_sl)
      @ [(3 + handler_sz + 3), state]
      @ (StackState.translate (3 + handler_sz + 3) body_sl)
      @ [(3 + handler_sz + 3 + body_sz), (StackState.generalize body_s handler_s)],
      (StackState.generalize body_s handler_s)
  | Mswitch (arg, consts, actions, default) ->
      let consts_len = Array.length consts in
      let arg_instrs, arg_sz, arg_st, arg_sl, arg_s' =
        compile_expression state stack false boxed_result ofs arg in
      let arg_s = StackState.pop arg_s' in
      let switch_ofs = ofs + arg_sz in
      let padding = 3 - (switch_ofs mod 4) in
      let switch_sz = 1 + padding + 4 + 4 + 4 + (4 * consts_len) in
      let o = ref (ofs + arg_sz + switch_sz) in
      let actions_sz = ref 0 in
      let max_st = ref 0 in
      let actions_len = Array.length actions in
      let compiled_actions = Array.make actions_len None in
      let frames = ref [] in
      let frames_after = ref [] in
      for i = 0 to pred actions_len do
        let act_instrs, act_sz, act_st, act_sl, act_s =
          compile_expression arg_s stack is_tail boxed_result !o actions.(i) in
        compiled_actions.(i) <- Some (act_instrs, !o, !o + act_sz);
        frames := !frames
          @ [(!o - ofs), arg_s]
          @ (StackState.translate (!o - ofs) act_sl)
          @ [(!o - ofs + act_sz), act_s];
        frames_after := act_s :: !frames_after;
        actions_sz := !actions_sz + act_sz + 3;
        max_st := max !max_st act_st;
        o := !o + act_sz + 3
      done;
      let default_instrs, default_sz, default_st, default_sl, default_s =
        compile_expression arg_s stack false boxed_result !o default in
      frames_after := default_s :: !frames_after;
      o := !o + default_sz;
      max_st := max !max_st default_st;
      arg_instrs
      @ [ Instruction.TABLESWITCH
            (Int32.of_int (!o - switch_ofs - default_sz),
             Int32.zero,
             Int32.of_int (pred consts_len),
             Array.to_list
               (Array.map
                  (fun idx ->
                    match compiled_actions.(idx) with
                    | Some (_, o, _) ->
                        Int32.of_int (o - switch_ofs)
                    | None -> Misc.fatal_error "Bytecodegen.compile_expression: error in switch compilation")
                  consts)) ]
      @ (List.concat
           (Array.to_list
              (Array.map
                 (function
                   | Some (instrs, _, ofs) ->
                       instrs
                       @ [ Instruction.GOTO (check_jump (!o - ofs)) ]
                   | None -> Misc.fatal_error "Bytecodegen.compile_expression: error in switch compilation")
                 compiled_actions)))
      @ default_instrs,
      arg_sz + switch_sz + !actions_sz + default_sz,
      max arg_st !max_st,
      arg_sl
      @ [(arg_sz + switch_sz + !actions_sz), arg_s]
      @ (StackState.translate (arg_sz + switch_sz + !actions_sz) default_sl)
      @ !frames
      @ (List.map (fun x -> (arg_sz + switch_sz + !actions_sz + default_sz, x)) !frames_after),
      default_s
  | Mint expr ->
      compile_expression state stack false Unboxed ofs expr
  | Mtag expr ->
      let expr_instrs, expr_sz, expr_st, expr_sl, _ =
        compile_expression state stack false Boxed ofs expr in
      expr_instrs
      @ meth_asBlock
      @ meth_getTag,
      expr_sz + meth_size + meth_size,
      expr_st,
      expr_sl,
      (StackState.push_int state)
and compile_expression_list state stack boxed_results ofs l =
  let ofs', l', st, _, _, stat, frames = List.fold_left
      (fun (o, l, st, nb, ste, stat, frames) elem ->
        let res, sz, st', sl, s =
          compile_expression stat ste false boxed_results o elem in
        (o + sz, res :: l, (max st (st' + nb + 1)), succ nb, boxed_results :: ste, s, (frames @ (StackState.translate (o - ofs) sl))))
      (ofs, [], 0, 0, stack, state, [])
      l in
  (List.concat (List.rev l'), ofs' - ofs, st, frames, stat)
and compile_expression_list_to_locals state stack ofs l =
  let instrs = ref [] in
  let instrs_sz = ref 0 in
  let stores = ref [] in
  let stores_sz = ref 0 in
  let max_st = ref 0 in
  let o = ref ofs in
  let st = ref stack in
  let idx = ref 1 in
  let curr = ref state in
  let frames = ref [] in
  List.iter
    (fun elem ->
      let elem_instrs, elem_sz, elem_st, elem_sl, elem_s =
        compile_expression !curr !st false Boxed !o elem in
      let store_instrs, store_sz, _ = astore !idx in
      instrs := !instrs @ elem_instrs;
      instrs_sz := !instrs_sz + elem_sz;
      stores := store_instrs @ !stores;
      stores_sz := !stores_sz + store_sz;
      max_st := max !max_st (!idx + elem_st);
      curr := StackState.store_value !idx elem_s;
      frames := !frames @ (StackState.translate (!o - ofs) elem_sl);
      o := !o + elem_sz;
      st := Boxed :: !st;
      incr idx)
    l;
  !instrs @ !stores, !o - ofs + !stores_sz, !max_st, !frames, state
and compile_block state stack tag mut ofs args =
  let size = List.length args in
  if size = 0 && mut = Asttypes.Immutable then
    let tag_instrs, tag_sz, tag_st = push_int size in
    [ Instruction.ALOAD_0 ]
    @ tag_instrs
    @ meth_getAtom,
    1 + tag_sz + meth_size,
    1 + tag_st,
    [],
    state
  else
    (match size with
    | 1 | 2 | 3 | 4 ->
        let args_instrs, args_sz, args_st, args_sl, _ =
          compile_expression_list state stack Boxed ofs args in
        let tag_instrs, tag_sz, tag_st = push_int tag in
        args_instrs
        @ tag_instrs
        @ [ Instruction.INVOKESTATIC
              (class_AbstractNativeRunner,
               make_method "createBlock",
               ((parameters size) @ [`Int], `Class class_Value)) ],
        args_sz + tag_sz + meth_size,
        max args_st (size + tag_st),
        args_sl,
        (StackState.push_value state)
    | _ ->
        let args_instrs, args_sz, args_st, args_sl, _ =
          compile_expression_list state stack Boxed ofs args in
        let tag_instrs, tag_sz, tag_st = push_int tag in
        let size_instrs, size_sz, size_st = push_int size in
        let fold_instrs = ref [] in
        let fold_sz = ref 0 in
        let remaining = ref size in
        while !remaining > 0 do
          let n = min 4 !remaining in
          fold_instrs := !fold_instrs
            @ [ Instruction.INVOKESTATIC
                  (class_AbstractNativeRunner,
                   make_method ("foldValues" ^ (string_of_int n)),
                   ((parameters n) @ [`Class class_Values],
                    `Class class_Values)) ];
          fold_sz := !fold_sz + meth_size;
          remaining := !remaining - n
        done;
        args_instrs
        @ tag_instrs
        @ size_instrs
        @ meth_createValues
        @ !fold_instrs
        @ meth_returnValues,
        args_sz + tag_sz + size_sz + meth_size + !fold_sz + meth_size,
        max args_st (size + tag_st + size_st),
        args_sl,
        (StackState.push_value state))
and compile_first_closure cls func arity =
  let load_instrs, load_sz, load_st = aload !garbage_index in
  (if arity = 1 then
    load_instrs
    @ [ Instruction.DUP ;
        Instruction.ALOAD_0 ;
        Instruction.LDC_W (`Class_or_interface (make_class cls)) ;
        Instruction.LDC_W (`String (normalize_function_name func)) ]
    @ meth_registerClosure
    @ meth_setCode
    @ [ Instruction.ICONST_1 ]
    @ field_ONE
    @ meth_set,
    load_sz + 1 + 1 + 3 + 3 + meth_size + meth_size + 1 + field_size + meth_size,
    load_st + 4,
    2
  else
    let lbl = if arity < 0 then -1 else -arity - 1 in
    let lbl_instrs, lbl_sz, lbl_st = push_int lbl in
    let arity_instrs, arity_sz, arity_st = push_boxed_int arity in
    load_instrs
    @ [ Instruction.DUP ;
        Instruction.DUP ]
    @ lbl_instrs
    @ meth_setCode
    @ [ Instruction.ICONST_1 ]
    @ arity_instrs
    @ meth_set
    @ [ Instruction.ICONST_2 ;
        Instruction.ALOAD_0 ;
        Instruction.LDC_W (`Class_or_interface (make_class cls)) ;
        Instruction.LDC_W (`String (normalize_function_name func)) ]
    @ meth_registerClosure
    @ meth_createFromLong
    @ meth_set,
    load_sz + 1 + 1 + lbl_sz + meth_size + 1 + arity_sz + meth_size + 1 + 1
      + 3 + 3 + meth_size + meth_size + meth_size,
    load_st + 2 + lbl_st + arity_st + 5,
    3)
and compile_other_closures next = function
  | [] -> [], 0, 0, next
  | others ->
      let load_instrs, load_sz, load_st = aload !garbage_index in
      let idx = ref next in
      let instrs = ref [] in
      let sz = ref 0 in
      let max_st = ref 0 in
      List.iter
        (fun ((cls, func), arity) ->
          let idx_instrs, idx_sz, idx_st = push_int !idx in
          if arity = 1 then begin
            instrs := !instrs
              @ [ Instruction.ALOAD_0 ]
              @ load_instrs
              @ idx_instrs
              @ [ Instruction.LDC_W (`Class_or_interface (make_class cls)) ;
                  Instruction.LDC_W (`String (normalize_function_name func)) ]
              @ meth_registerInfix1;
            sz := !sz + 1 + load_sz + idx_sz + 3 + 3 + meth_size;
            idx := !idx + 3;
            let m = max 5 (max (1 + load_st) (2 + idx_st)) in
            max_st := max !max_st m
          end else begin
            let arity_instrs, arity_sz, arity_st = push_int arity in
            instrs := !instrs
              @ [ Instruction.ALOAD_0 ]
              @ load_instrs
              @ idx_instrs
              @ [ Instruction.LDC_W (`Class_or_interface (make_class cls)) ;
                  Instruction.LDC_W (`String (normalize_function_name func)) ]
              @ arity_instrs
              @ meth_registerInfixN;
            sz := !sz + 1 + load_sz + idx_sz + 3 + 3 + arity_sz + meth_size;
            idx := !idx + 4;
            let m = max 6 (max (1 + load_st) (max (2 + idx_st) (5 + arity_st))) in
            max_st := max !max_st m
          end)
        others;
      !instrs, !sz, !max_st, !idx

let compile_class meths =
  let classname = !Jclflags.java_package
    ^ "." ^ (Jcompilenv.current_unit_name ()) in
  let source_file =
    String.uncapitalize (Jcompilenv.current_unit_name ()) ^ ".ml" in
  let cd = { ClassDefinition.access_flags = [`Public; `Final; `Super];
             name = make_class classname;
             extends = Some class_Object;
             implements = [];
             fields = [];
             methods = meths;
             attributes = [`SourceFile (utf8_of_string source_file)]; } in
  let buff = Buffer.create 16 in
  let os = OutputStream.make_of_buffer buff in
  (try
    ClassFile.write (ClassDefinition.encode cd) os;
  with (BaristaLibrary.Attribute.Exception BaristaLibrary.Attribute.Invalid_code_length) ->
    raise (Error (Method_too_long classname)));
  OutputStream.close os;
  Buffer.contents buff

let compile_function fd =
  let rec signature acc = function
    | 0 -> (`Class class_AbstractNativeRunner) :: acc
    | n -> signature ((`Class class_Value) :: acc) (pred n) in
  current_file :=
    String.uncapitalize (Jcompilenv.current_unit_name ()) ^ ".ml";
  current_class := !Jclflags.java_package
    ^ "." ^ (Jcompilenv.current_unit_name ());
  current_function := fd.fun_name;
  exception_table := [];
  garbage_index := fd.fun_garbage_index;
  Hashtbl.clear current_catches;
  line_number_table := [];
  prev_ofs := 0;
  let parameters = signature [] (List.length fd.fun_args) in
  initial_state := StackState.make fd.fun_max_locals parameters;
  let instrs, size, max_stack, frames, state =
    compile_expression !initial_state [] true Boxed 0 fd.fun_body in
  let frames' = StackState.to_stackmap_list !initial_state (frames @ [size, state]) in
  let code = { Attribute.max_stack = max_stack;
               Attribute.max_locals = fd.fun_max_locals;
               Attribute.code = instrs @ [ Instruction.ARETURN ];
               Attribute.exception_table = List.rev_map
                 (fun (start_pc, end_pc, handler_pc) ->
                   (start_pc, end_pc, handler_pc, (Some class_FailException)))
                 !exception_table;
               Attribute.attributes =
               (if frames' <> [] then
                 [ `StackMapTable frames' ]
               else
                 [])
               @ (if !line_number_table <> [] then
                 [ `LineNumberTable (List.rev !line_number_table) ]
               else
                 [])
               @ (if !Clflags.debug then
                 [ `LocalVariableTable
                     ((0,
                       size,
                       utf8_of_string "runner",
                       `Class class_AbstractNativeRunner,
                       0)
                      :: (List.rev
                            (snd (List.fold_left
                                    (fun (idx, l) id ->
                                      succ idx,
                                      ((0,
                                        size,
                                        (utf8_of_string (Ident.name id)),
                                        (`Class class_Value),
                                        idx) :: l))
                                    (1, [])
                                    fd.fun_args)))) ]
               else
                 []) ; } in
  Method.Regular ([`Public; `Static],
                  make_method fd.fun_name,
                  (parameters, `Class class_Value),
                  [`Code code;
                   `Exceptions [class_FailException; class_FatalException]])

let report_error ppf = function
  | Method_too_long name ->
      Format.fprintf ppf "Cannot compile %s (Java method is too long)" name
