(*
 * This file is part of Cafesterol.
 * Copyright (C) 2007-2009 Xavier Clerc.
 *
 * Cafesterol is free software; you can redistribute it and/or modify
 * it under the terms of the Q Public License as published by
 * Trolltech (with a change to choice of law).
 *
 * Cafesterol is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Q Public License for more details.
 *
 * You should have received a copy of the Q Public License
 * along with this program.  If not, see
 * <http://trolltech.com/products/qt/licenses/licensing/qpl>.
 *)

open Lambda
open BaristaLibrary
open Utils
open Bytecodeutils


exception Not_a_call

let as_mcall prim args =
  let make_call prim_name args =
    let class_name = Cadmiumprimitives.get_class prim_name in
    Macroinstr.Mcall (Macroinstr.Primitive,
                      (class_name, prim_name),
                      args,
                      Debuginfo.none) in
  let make_bint_call bi suff args =
    let pref = match bi with
    | Pnativeint -> "caml_nativeint_"
    | Pint32 -> "caml_int32_"
    | Pint64 -> "caml_int64_" in
    make_call (pref ^ suff) args in
  match prim with
  | Pidentity -> raise Not_a_call
  | Pignore -> raise Not_a_call
  | Pgetglobal _ -> raise Not_a_call
  | Psetglobal _ -> raise Not_a_call
  | Pmakeblock _ -> raise Not_a_call
  | Pfield _ -> raise Not_a_call
  | Psetfield _ -> raise Not_a_call
  | Pfloatfield _ -> raise Not_a_call
  | Psetfloatfield _ -> raise Not_a_call
  | Pduprecord _ -> make_call "caml_obj_dup" args
  | Pccall p -> make_call p.Primitive.prim_name args
  | Praise -> raise Not_a_call
  | Psequand | Psequor | Pnot -> raise Not_a_call
  | Pnegint | Paddint | Psubint | Pmulint | Pdivint | Pmodint
  | Pandint | Porint | Pxorint
  | Plslint | Plsrint | Pasrint
  | Pintcomp _
  | Poffsetint _
  | Poffsetref _ -> raise Not_a_call
  | Pintoffloat -> make_call "caml_int_of_float" args
  | Pfloatofint -> make_call "caml_float_of_int" args
  | Pnegfloat -> make_call "caml_neg_float" args
  | Pabsfloat -> make_call "caml_abs_float" args
  | Paddfloat -> make_call "caml_add_float" args
  | Psubfloat -> make_call "caml_sub_float" args
  | Pmulfloat -> make_call "caml_mul_float" args
  | Pdivfloat -> make_call "caml_div_float" args
  | Pfloatcomp Ceq -> make_call "caml_eq_float" args
  | Pfloatcomp Cneq -> make_call "caml_neq_float" args
  | Pfloatcomp Clt -> make_call "caml_lt_float" args
  | Pfloatcomp Cgt -> make_call "caml_gt_float" args
  | Pfloatcomp Cle -> make_call "caml_le_float" args
  | Pfloatcomp Cge -> make_call "caml_ge_float" args
  | Pstringlength -> make_call "caml_ml_string_length" args
  | Pstringrefu -> raise Not_a_call
  | Pstringsetu -> raise Not_a_call
  | Pstringrefs -> make_call "caml_string_get" args
  | Pstringsets -> make_call "caml_string_set" args
  | Pmakearray _ -> raise Not_a_call
  | Parraylength _ -> raise Not_a_call
  | Parrayrefu Pgenarray -> make_call "caml_array_unsafe_get" args
  | Parrayrefu Pfloatarray -> make_call "caml_array_unsafe_get_float" args
  | Parrayrefu _ -> raise Not_a_call
  | Parraysetu Pgenarray -> make_call "caml_array_unsafe_set" args
  | Parraysetu Pfloatarray -> make_call "caml_array_unsafe_set_float" args
  | Parraysetu _ -> raise Not_a_call
  | Parrayrefs Pgenarray -> make_call "caml_array_get" args
  | Parrayrefs Pfloatarray -> make_call "caml_array_get_float" args
  | Parrayrefs _ -> make_call "caml_array_get_addr" args
  | Parraysets Pgenarray -> make_call "caml_array_set" args
  | Parraysets Pfloatarray -> make_call "caml_array_set_float" args
  | Parraysets _ -> make_call "caml_array_set_addr" args
  | Pisint -> raise Not_a_call
  | Pisout -> raise Not_a_call
  | Pbittest -> make_call "caml_bitvect_test" args
  | Pbintofint bi -> make_bint_call bi "of_int" args
  | Pintofbint bi -> make_bint_call bi "to_int" args
  | Pcvtbint(Pint32, Pnativeint) -> make_call "caml_nativeint_of_int32" args
  | Pcvtbint(Pnativeint, Pint32) -> make_call "caml_nativeint_to_int32" args
  | Pcvtbint(Pint32, Pint64) -> make_call "caml_int64_of_int32" args
  | Pcvtbint(Pint64, Pint32) -> make_call "caml_int64_to_int32" args
  | Pcvtbint(Pnativeint, Pint64) -> make_call "caml_int64_of_nativeint" args
  | Pcvtbint(Pint64, Pnativeint) -> make_call "caml_int64_to_nativeint" args
  | Pnegbint bi -> make_bint_call bi "neg" args
  | Paddbint bi -> make_bint_call bi "add" args
  | Psubbint bi -> make_bint_call bi "sub" args
  | Pmulbint bi -> make_bint_call bi "mul" args
  | Pdivbint bi -> make_bint_call bi "div" args
  | Pmodbint bi -> make_bint_call bi "mod" args
  | Pandbint bi -> make_bint_call bi "and" args
  | Porbint bi -> make_bint_call bi "or" args
  | Pxorbint bi -> make_bint_call bi "xor" args
  | Plslbint bi -> make_bint_call bi "shift_left" args
  | Plsrbint bi -> make_bint_call bi "shift_right_unsigned" args
  | Pasrbint bi -> make_bint_call bi "shift_right" args
  | Pbintcomp (_, Ceq) -> make_call "caml_equal" args
  | Pbintcomp (_, Cneq) -> make_call "caml_notequal" args
  | Pbintcomp (_, Clt) -> make_call "caml_lessthan" args
  | Pbintcomp (_, Cgt) -> make_call "caml_greaterthan" args
  | Pbintcomp (_, Cle) -> make_call "caml_lessequal" args
  | Pbintcomp (_, Cge) -> make_call "caml_greaterequal" args
  | Pbigarrayref (_, n, _, _) -> make_call ("caml_ba_get_" ^ string_of_int n) args
  | Pbigarrayset (_, n, _, _) -> make_call ("caml_ba_set_" ^ string_of_int n) args
  | _ -> Misc.fatal_error "Bytecodeprimitive.as_mcall: unhandled primitive"

let is_mcall prim =
  try
    ignore (as_mcall prim []); true
  with Not_a_call -> false

let parameters_type = function
  | Pgetglobal _ -> Macroinstr.Boxed
  | Psetglobal _ -> Macroinstr.Boxed
  | Pfield _ -> Macroinstr.Boxed
  | Psetfield _ -> Macroinstr.Boxed
  | Pfloatfield _ -> Macroinstr.Boxed
  | Psetfloatfield _ -> Macroinstr.Boxed
  | Praise -> Macroinstr.Boxed
  | Pnegint | Paddint | Psubint | Pmulint | Pdivint | Pmodint
  | Pandint | Porint | Pxorint
  | Plslint | Plsrint | Pasrint
  | Poffsetint _ -> Macroinstr.Unboxed
  | Poffsetref _ -> Macroinstr.Boxed
  | Parraylength _ -> Macroinstr.Boxed
  | _ ->
      Misc.fatal_error "Bytecodeprimitive.parameters_type: unhandled primitive"

let compile state = function
  | Pgetglobal id ->
      [ Instruction.ALOAD_0 ;
        Instruction.LDC_W (`String (utf8_of_string (Ident.name id))) ]
      @ meth_getGlobal,
      1 + 3 + meth_size,
      2,
      []
  | Psetglobal id ->
      [ Instruction.ALOAD_0 ;
        Instruction.SWAP ;
        Instruction.LDC_W (`String (utf8_of_string (Ident.name id))) ;
        Instruction.SWAP ]
      @ meth_setGlobal,
      1 + 1 + 3 + 1 + meth_size,
      2,
      []
  | Pfield n ->
      let int_instrs, int_sz, int_st = push_int n in
      meth_asBlock
      @ int_instrs
      @ meth_get,
      meth_size + int_sz + meth_size,
      int_st,
      []
  | Psetfield (n, _) ->
      let int_instrs, int_sz, int_st = push_int n in
      [ Instruction.SWAP ]
      @ meth_asBlock
      @ [ Instruction.SWAP ]
      @ int_instrs
      @ [ Instruction.SWAP ]
      @ meth_set
      @ field_UNIT,
      1 + meth_size + 1 + int_sz + 1 + meth_size + field_size,
      int_st,
      []
  | Pfloatfield n ->
      let int_instrs, int_sz, int_st = push_int n in
      meth_asBlock
      @ [ Instruction.DUP ]
      @ meth_getTag
      @ field_DOUBLE_ARRAY_TAG
      @ [ Instruction.IF_ICMPNE (3 + int_sz + meth_size + 3) ]
      @ int_instrs
      @ meth_getDouble
      @ [ Instruction.GOTO (3 + int_sz + (meth_size * 3)) ]
      @ int_instrs
      @ meth_get
      @ meth_asBlock
      @ meth_asDouble
      @ meth_createDouble
      @ meth_createFromBlock,
      meth_size + 1 + meth_size + field_size + 3 + int_sz + meth_size + 3
        + int_sz + (5 * meth_size),
      max 2 int_st,
      [(meth_size + 1 + meth_size + field_size + 3 + int_sz + meth_size + 3,
        (StackState.push_block (StackState.pop state)));
       (meth_size + 1 + meth_size + field_size + 3 + int_sz + meth_size + 3 + int_sz + (3 * meth_size),
       (StackState.push_double (StackState.pop state)))]
  | Psetfloatfield n ->
      let int_instrs, int_sz, int_st = push_int n in
      [ Instruction.SWAP ]
      @ meth_asBlock
      @ [ Instruction.DUP ]
      @ meth_getTag
      @ field_DOUBLE_ARRAY_TAG
      @ [ Instruction.IF_ICMPEQ (3 + 1 + int_sz + 1 + (5 * meth_size) + 3) ;
          Instruction.SWAP ]
      @ int_instrs
      @ [ Instruction.SWAP ]
      @ meth_asBlock
      @ meth_asDouble
      @ meth_createDouble
      @ meth_createFromBlock
      @ meth_set
      @ [ Instruction.GOTO (3 + 1 + int_sz + 1 + (3 * meth_size)) ;
          Instruction.SWAP ]
      @ int_instrs
      @ [ Instruction.SWAP ]
      @ meth_asBlock
      @ meth_asDouble
      @ meth_setDouble
      @ field_UNIT,
      1 + meth_size + 1 + meth_size + field_size + 3 + 1 + int_sz + 1
        + (5 * meth_size) + 3 + 1 + int_sz + 1 + (3 * meth_size) + field_size,
      max 2 int_st,
      [(1 + meth_size + 1 + meth_size + field_size + 3 + 1 + int_sz + 1 + (5 * meth_size) + 3,
        (StackState.push_block (StackState.pop state)));
       (1 + meth_size + 1 + meth_size + field_size + 3 + 1 + int_sz + 1 + (5 * meth_size) + 3 + 1 + int_sz + 1 + (3 * meth_size),
        (StackState.pop (StackState.pop state)))]
  | Praise ->
      meth_raise
      @ field_UNIT,
      meth_size + field_size,
      0,
      []
  | Pnegint ->
      [ Instruction.INEG ], 1, 0, []
  | Paddint ->
      [ Instruction.IADD ], 1, 0, []
  | Psubint ->
      [ Instruction.ISUB ], 1, 0, []
  | Pmulint ->
      [ Instruction.IMUL ], 1, 0, []
  | Pdivint ->
      [ Instruction.DUP ;
        Instruction.IFEQ (3 + 1 + 3) ;
        Instruction.IDIV ;
        Instruction.GOTO (3 + 1 + meth_size) ;
        Instruction.POP ]
      @ meth_raiseZeroDivide,
      1 + 3 + 1 + 3 + 1 + meth_size,
      1,
      [(1 + 3 + 1 + 3, StackState.same state);
       (1 + 3 + 1 + 3 + 1 + meth_size, StackState.pop state)]
  | Pmodint ->
      [ Instruction.DUP ;
        Instruction.IFEQ (3 + 1 + 3) ;
        Instruction.IREM ;
        Instruction.GOTO (3 + 1 + meth_size) ;
        Instruction.POP ]
      @ meth_raiseZeroDivide,
      1 + 3 + 1 + 3 + 1 + meth_size,
      1,
      [(1 + 3 + 1 + 3, StackState.same state);
       (1 + 3 + 1 + 3 + 1 + meth_size, StackState.pop state)]
  | Pandint ->
      [ Instruction.IAND ], 1, 0, []
  | Porint ->
      [ Instruction.IOR ], 1, 0, []
  | Pxorint ->
      [ Instruction.IXOR ], 1, 0, []
  | Plslint ->
      [ Instruction.ISHL ], 1, 0, []
  | Plsrint ->
      [ Instruction.IUSHR ], 1, 0, []
  | Pasrint ->
      [ Instruction.ISHR ], 1, 0, []
  | Poffsetint n ->
      let int_instrs, int_sz, int_st = push_int n in
      int_instrs @ [ Instruction.IADD ], int_sz + 1, int_st, []
  | Poffsetref n ->
      let int_instrs, int_sz, int_st = push_int (n * 2) in
      meth_asBlock
      @ [ Instruction.DUP ;
          Instruction.ICONST_0 ]
      @ meth_get
      @ meth_getRawValue
      @ int_instrs
      @ [ Instruction.IADD ]
      @ meth_createFromRawValue
      @ [ Instruction.ICONST_0 ;
          Instruction.SWAP ]
      @ meth_set
      @ field_UNIT,
      meth_size + 1 + 1 + meth_size + meth_size + int_sz + 1 + meth_size + 1
        + 1 + meth_size + field_size,
      2 + int_st + 1,
      []
  | Parraylength _ ->
      meth_asBlock
      @ [ Instruction.DUP ]
      @ meth_getWoSize
      @ [ Instruction.SWAP ]
      @ meth_getTag
      @ field_DOUBLE_ARRAY_TAG
      @ [ Instruction.IF_ICMPNE (3 + field_size + 1) ]
      @ field_DOUBLE_SIZE
      @ [ Instruction.IDIV ]
      @ meth_createFromLong,
      meth_size + 1 + meth_size + 1 + meth_size + field_size + 3 + field_size
        + 1 + meth_size,
      2,
      [(meth_size + 1 + meth_size + 1 + meth_size + field_size + 3 + field_size + 1,
        (StackState.push_int (StackState.pop state)))]
  | _ -> Misc.fatal_error "Bytecodeprimitive.compile: unhandled primitive"
