(*
 * This file is part of Cafesterol.
 * Copyright (C) 2007-2009 Xavier Clerc.
 * Original file (asmcomp/asmpackager.ml in the Objective Caml source
 * distribution) is Copyright (C) INRIA.
 *
 * Cafesterol is free software; you can redistribute it and/or modify
 * it under the terms of the Q Public License as published by
 * Trolltech (with a change to choice of law).
 *
 * Cafesterol is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Q Public License for more details.
 *
 * You should have received a copy of the Q Public License
 * along with this program.  If not, see
 * <http://trolltech.com/products/qt/licenses/licensing/qpl>.
 *)

open Jcompilenv


type error =
  | Illegal_renaming of string * string
  | Forward_reference of string * string
  | Wrong_for_pack of string * string
  | Linking_error
  | File_not_found of string

exception Error of error


type pack_member_kind = PM_intf | PM_impl of unit_infos

type pack_member =
    { pm_file : string;
      pm_name : string;
      pm_kind : pack_member_kind; }

let read_member_info pack_path file =
  let name = String.capitalize (Filename.basename (Misc.chop_extensions file)) in
  let kind =
    if Filename.check_suffix file ".cmj" then begin
      let info, crc = read_unit_info file in
      if info.ui_name <> name then
        raise (Error (Illegal_renaming (file, info.ui_name)));
      if info.ui_symbol <>
        (current_unit_infos ()).ui_symbol
        ^ "__" ^ info.ui_name then
        raise (Error (Wrong_for_pack (file, pack_path)));
      Javalink.check_consistency file info crc;
      cache_unit_info info;
      PM_impl info
    end else
      PM_intf in
  { pm_file = file; pm_name = name; pm_kind = kind; }

let check_units members =
  let rec check forbidden = function
    | [] -> ()
    | mb :: tl ->
        (match mb.pm_kind with
        | PM_intf -> ()
        | PM_impl infos ->
            List.iter
              (fun (unit, _) ->
                if List.mem unit forbidden then
                  raise (Error (Forward_reference (mb.pm_file, unit))))
              infos.ui_imports_cmj);
        check (Misc.list_remove mb.pm_name forbidden) tl in
  check (List.map (fun mb -> mb.pm_name) members) members

let make_package_object ppf members targetobj targetname coercion =
  let objtemp = Filename.temp_file (make_symbol (Some "")) Jconfig.ext_obj in
  let components = List.map
      (fun m ->
      match m.pm_kind with
      | PM_intf -> None
      | PM_impl _ -> Some (Ident.create_persistent m.pm_name))
      members in
  Javagen.compile_implementation
    (Misc.chop_extension_if_any objtemp)
    ppf
    (Translmod.transl_store_package
       components
       (Ident.create_persistent targetname)
       coercion);
  let objfiles =
    List.map
      (fun m -> Misc.chop_extension_if_any m.pm_file ^ Jconfig.ext_obj)
      (List.filter (fun m -> m.pm_kind <> PM_intf) members) in
  (try
    let zip = Zip.open_out targetobj in
    List.iter
      (fun fn ->
        let z = Zip.open_in fn in
        List.iter
          (fun e ->
            if not e.Zip.is_directory then
              let data = Zip.read_entry z e in
              Zip.add_entry data zip e.Zip.filename)
          (Zip.entries z);
        Zip.close_in z)
      (objtemp :: objfiles);
    Zip.close_out zip;
  with
  | _ -> raise (Error Linking_error));
  Misc.remove_file objtemp

let build_package_cmj members cmjfile =
  let unit_names = List.map (fun m -> m.pm_name) members in
  let filter lst =
    List.filter (fun (name, _) -> not (List.mem name unit_names)) lst in
  let units = List.fold_right
      (fun m acc ->
        match m.pm_kind with
        | PM_intf -> acc
        | PM_impl info -> info :: acc)
      members
      [] in
  let ui = current_unit_infos () in
  let pkg_infos =
    { ui_name = ui.ui_name;
      ui_javaname = ui.ui_javaname;
      ui_additional_classes = List.flatten
        (List.map
           (fun info ->
             info.ui_additional_classes @ [info.ui_javaname])
           units);
      ui_symbol = ui.ui_symbol;
      ui_defines =
      List.flatten (List.map (fun info -> info.ui_defines) units)
      @ [ui.ui_symbol] ;
      ui_imports_cmi = (ui.ui_name, Env.crc_of_unit ui.ui_name) ::
      (filter (Javalink.extract_crc_interfaces ()));
      ui_imports_cmj = filter (Javalink.extract_crc_implementations ());
      ui_approx = ui.ui_approx; } in
  write_unit_info pkg_infos cmjfile

let package_object_files ppf files targetcmj targetobj targetname coercion =
  let pack_path = match !Clflags.for_package with
  | None -> targetname
  | Some p -> p ^ "." ^ targetname in
  let members = Misc.map_left_right (read_member_info pack_path) files in
  check_units members;
  make_package_object ppf members targetobj targetname coercion;
  build_package_cmj members targetcmj

let package_files ppf files targetcmj =
  let files = List.map
      (fun f ->
        try
          Misc.find_in_path !Config.load_path f
        with
        | Not_found -> raise (Error (File_not_found f)))
      files in
  let prefix = Misc.chop_extensions targetcmj in
  let targetcmi = prefix ^ ".cmi" in
  let targetobj = Misc.chop_extension_if_any targetcmj ^ Jconfig.ext_obj in
  let targetname = String.capitalize (Filename.basename prefix) in
  Location.input_name := targetcmj;
  reset ?packname:!Clflags.for_package targetname;
  try
    let coercion = Typemod.package_units files targetcmi targetname in
    package_object_files ppf files targetcmj targetobj targetname coercion
  with x ->
    Misc.remove_file targetcmj;
    Misc.remove_file targetobj;
    raise x

let report_error ppf = function
  | Illegal_renaming(file, id) ->
      Format.fprintf ppf "Wrong file naming: %s@ contains the code for@ %s"
        file id
  | Forward_reference(file, ident) ->
      Format.fprintf ppf "Forward reference to %s in file %s" ident file
  | Wrong_for_pack(file, path) ->
      Format.fprintf ppf "File %s@ was not compiled with the `-for-pack %s' option"
              file path
  | File_not_found file ->
      Format.fprintf ppf "File %s not found" file
  | Linking_error ->
      Format.fprintf ppf "Error during partial linking"
