(*
 * This file is part of Cafesterol.
 * Copyright (C) 2007-2009 Xavier Clerc.
 * Original file (asmcomp/compilenv.ml in the Objective Caml source
 * distribution) is Copyright (C) INRIA.
 *
 * Cafesterol is free software; you can redistribute it and/or modify
 * it under the terms of the Q Public License as published by
 * Trolltech (with a change to choice of law).
 *
 * Cafesterol is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Q Public License for more details.
 *
 * You should have received a copy of the Q Public License
 * along with this program.  If not, see
 * <http://trolltech.com/products/qt/licenses/licensing/qpl>.
 *)

type error =
  | Not_a_unit_info of string
  | Corrupted_unit_info of string
  | Illegal_renaming of string* string

exception Error of error


type unit_infos =
    { mutable ui_name : string;
      mutable ui_javaname : string;
      mutable ui_additional_classes : string list;
      mutable ui_symbol : string;
      mutable ui_defines : string list;
      mutable ui_imports_cmi : (string * Digest.t) list;
      mutable ui_imports_cmj : (string * Digest.t) list;
      mutable ui_approx : Jlambda.value_approximation; }

let cmj_not_found_crc = String.make 16 '\000'

type library_infos =
    { lib_units : (unit_infos * Digest.t) list;
      lib_ccobjs : string list; }

let global_infos_table : (string, unit_infos option) Hashtbl.t =
  Hashtbl.create 17

let current_unit =
    { ui_name = "";
      ui_javaname = "";
      ui_additional_classes = [];
      ui_symbol = "";
      ui_defines = [];
      ui_imports_cmi = [];
      ui_imports_cmj = [];
      ui_approx = Jlambda.Value_unknown; }

let symbolname_for_pack pack name =
  match pack with
  | None -> name
  | Some p ->
      let b = Buffer.create 64 in
      for i = 0 to pred (String.length p) do
        match p.[i] with
        | '.' -> Buffer.add_string b "__"
        | c -> Buffer.add_char b c
      done;
      Buffer.add_string b "__";
      Buffer.add_string b name;
      Buffer.contents b

let reset ?packname name =
  Hashtbl.clear global_infos_table;
  let symbol = symbolname_for_pack packname name in
  current_unit.ui_name <- name;
  current_unit.ui_javaname <- !Jclflags.java_package ^ "." ^ name;
  current_unit.ui_additional_classes <- [];
  current_unit.ui_symbol <- symbol;
  current_unit.ui_defines <- [symbol];
  current_unit.ui_imports_cmi <- [];
  current_unit.ui_imports_cmj <- [];
  current_unit.ui_approx <- Jlambda.Value_unknown

let current_unit_infos () =
  current_unit

let current_unit_name () =
  current_unit.ui_name

let read_unit_info filename =
  let ic = open_in_bin filename in
  try
    let len = String.length Jconfig.cmj_magic_number in
    let buffer = String.create len in
    really_input ic buffer 0 len;
    if buffer <> Jconfig.cmj_magic_number then begin
      close_in ic;
      raise (Error (Not_a_unit_info filename))
    end;
    let ui : unit_infos = input_value ic in
    let crc = Digest.input ic in
    close_in ic;
    ui, crc
  with
  | End_of_file
  | Failure _ ->
      close_in ic;
      raise (Error (Corrupted_unit_info filename))

let write_unit_info info filename =
  let oc = open_out_bin filename in
  output_string oc Jconfig.cmj_magic_number;
  output_value oc info;
  flush oc;
  let crc = Digest.file filename in
  Digest.output oc crc;
  close_out oc

let save_unit_info filename =
  current_unit.ui_imports_cmi <- Env.imported_units ();
  write_unit_info current_unit filename

let cache_unit_info ui =
  Hashtbl.add global_infos_table ui.ui_name (Some ui)

let make_symbol ?(unitname = current_unit.ui_symbol) idopt =
  match idopt with
  | None -> unitname
  | Some id -> id

let get_global_info global_ident =
  let modname = Ident.name global_ident in
  if modname = current_unit.ui_name then
    Some current_unit
  else begin
    try
      Hashtbl.find global_infos_table modname
    with Not_found ->
      let infos, crc = try
        let filename = Misc.find_in_path_uncap !Config.load_path (modname ^".cmj") in
        let ui, crc = read_unit_info filename in
        if ui.ui_name <> modname then
          raise (Error (Illegal_renaming (ui.ui_name, filename)));
        Some ui, crc
      with Not_found ->
        None, cmj_not_found_crc in
      current_unit.ui_imports_cmj <- (modname, crc) :: current_unit.ui_imports_cmj;
      Hashtbl.add global_infos_table modname infos;
      infos
  end

let symbol_for_global id =
  if Ident.is_predef_exn id then
    "caml_exn_" ^ Ident.name id
  else begin
    match get_global_info id with
    | None -> make_symbol ~unitname:(Ident.name id) None
    | Some ui -> make_symbol ~unitname:ui.ui_symbol None
  end

let toplevel_approx = Hashtbl.create 16

let record_global_approx_toplevel () =
  Hashtbl.add toplevel_approx current_unit.ui_name current_unit.ui_approx

let global_approx id =
  if Ident.is_predef_exn id then Jlambda.Value_unknown
  else try Hashtbl.find toplevel_approx (Ident.name id)
  with Not_found ->
    match get_global_info id with
    | None -> Jlambda.Value_unknown
    | Some ui -> ui.ui_approx

let set_global_approx approx =
  current_unit.ui_approx <- approx

let read_library_info filename =
  let ic = open_in_bin filename in
  let len = String.length Jconfig.cmja_magic_number in
  let buffer = String.create len in
  really_input ic buffer 0 len;
  if buffer <> Jconfig.cmja_magic_number then
    raise (Error (Not_a_unit_info filename));
  let infos : library_infos = input_value ic in
  close_in ic;
  infos

let report_error ppf = function
  | Not_a_unit_info filename ->
      Format.fprintf ppf "%s@ is not a compilation unit description" filename
  | Corrupted_unit_info filename ->
      Format.fprintf ppf "Corrupted compilation unit description@ %s" filename
  | Illegal_renaming (modname, filename) ->
      Format.fprintf ppf "%s@ contains the description for unit@ %s" filename modname
