(*
 * This file is part of Cafesterol.
 * Copyright (C) 2007-2009 Xavier Clerc.
 *
 * Cafesterol is free software; you can redistribute it and/or modify
 * it under the terms of the Q Public License as published by
 * Trolltech (with a change to choice of law).
 *
 * Cafesterol is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Q Public License for more details.
 *
 * You should have received a copy of the Q Public License
 * along with this program.  If not, see
 * <http://trolltech.com/products/qt/licenses/licensing/qpl>.
 *)

open Format
open Jlambda
open Lambda
open Asttypes
open Types
open Primitive


let boxed_integer_name = function
  | Pnativeint -> "nativeint"
  | Pint32 -> "int32"
  | Pint64 -> "int64"

let print_boxed_integer name ppf bi =
  fprintf ppf "%s_%s" (boxed_integer_name bi) name

let print_boxed_integer_conversion ppf bi1 bi2 =
  fprintf ppf "%s_of_%s" (boxed_integer_name bi2) (boxed_integer_name bi1)

let boxed_integer_mark name = function
  | Pnativeint -> Printf.sprintf "Nativeint.%s" name
  | Pint32 -> Printf.sprintf "Int32.%s" name
  | Pint64 -> Printf.sprintf "Int64.%s" name

let print_boxed_integer name ppf bi =
  fprintf ppf "%s" (boxed_integer_mark name bi);;

let print_bigarray name unsafe kind ppf layout =
  fprintf ppf "Bigarray.%s[%s,%s]"
    (if unsafe then "unsafe_" ^ name else name)
    (match kind with
    | Pbigarray_unknown -> "generic"
    | Pbigarray_float32 -> "float32"
    | Pbigarray_float64 -> "float64"
    | Pbigarray_sint8 -> "sint8"
    | Pbigarray_uint8 -> "uint8"
    | Pbigarray_sint16 -> "sint16"
    | Pbigarray_uint16 -> "uint16"
    | Pbigarray_int32 -> "int32"
    | Pbigarray_int64 -> "int64"
    | Pbigarray_caml_int -> "camlint"
    | Pbigarray_native_int -> "nativeint"
    | Pbigarray_complex32 -> "complex32"
    | Pbigarray_complex64 -> "complex64")
    (match layout with
    |  Pbigarray_unknown_layout -> "unknown"
    | Pbigarray_c_layout -> "C"
    | Pbigarray_fortran_layout -> "Fortran")

let record_rep ppf r =
  match r with
  | Record_regular -> fprintf ppf "regular"
  | Record_float -> fprintf ppf "float"

let primitive ppf = function
  | Pidentity -> fprintf ppf "id"
  | Pignore -> fprintf ppf "ignore"
  | Pgetglobal id -> fprintf ppf "global %a" Ident.print id
  | Psetglobal id -> fprintf ppf "setglobal %a" Ident.print id
  | Pmakeblock(tag, Immutable) -> fprintf ppf "makeblock %i" tag
  | Pmakeblock(tag, Mutable) -> fprintf ppf "makemutable %i" tag
  | Pfield n -> fprintf ppf "field %i" n
  | Psetfield(n, ptr) ->
      let instr = if ptr then "setfield_ptr " else "setfield_imm " in
      fprintf ppf "%s%i" instr n
  | Pfloatfield n -> fprintf ppf "floatfield %i" n
  | Psetfloatfield n -> fprintf ppf "setfloatfield %i" n
  | Pduprecord (rep, size) -> fprintf ppf "duprecord %a %i" record_rep rep size
  | Plazyforce -> fprintf ppf "force"
  | Pccall p -> fprintf ppf "%s" p.prim_name
  | Praise -> fprintf ppf "raise"
  | Psequand -> fprintf ppf "&&"
  | Psequor -> fprintf ppf "||"
  | Pnot -> fprintf ppf "not"
  | Pnegint -> fprintf ppf "~"
  | Paddint -> fprintf ppf "+"
  | Psubint -> fprintf ppf "-"
  | Pmulint -> fprintf ppf "*"
  | Pdivint -> fprintf ppf "/"
  | Pmodint -> fprintf ppf "mod"
  | Pandint -> fprintf ppf "and"
  | Porint -> fprintf ppf "or"
  | Pxorint -> fprintf ppf "xor"
  | Plslint -> fprintf ppf "lsl"
  | Plsrint -> fprintf ppf "lsr"
  | Pasrint -> fprintf ppf "asr"
  | Pintcomp(Ceq) -> fprintf ppf "=="
  | Pintcomp(Cneq) -> fprintf ppf "!="
  | Pintcomp(Clt) -> fprintf ppf "<"
  | Pintcomp(Cle) -> fprintf ppf "<="
  | Pintcomp(Cgt) -> fprintf ppf ">"
  | Pintcomp(Cge) -> fprintf ppf ">="
  | Poffsetint n -> fprintf ppf "%i+" n
  | Poffsetref n -> fprintf ppf "+:=%i"n
  | Pintoffloat -> fprintf ppf "int_of_float"
  | Pfloatofint -> fprintf ppf "float_of_int"
  | Pnegfloat -> fprintf ppf "~."
  | Pabsfloat -> fprintf ppf "abs."
  | Paddfloat -> fprintf ppf "+."
  | Psubfloat -> fprintf ppf "-."
  | Pmulfloat -> fprintf ppf "*."
  | Pdivfloat -> fprintf ppf "/."
  | Pfloatcomp(Ceq) -> fprintf ppf "==."
  | Pfloatcomp(Cneq) -> fprintf ppf "!=."
  | Pfloatcomp(Clt) -> fprintf ppf "<."
  | Pfloatcomp(Cle) -> fprintf ppf "<=."
  | Pfloatcomp(Cgt) -> fprintf ppf ">."
  | Pfloatcomp(Cge) -> fprintf ppf ">=."
  | Pstringlength -> fprintf ppf "string.length"
  | Pstringrefu -> fprintf ppf "string.unsafe_get"
  | Pstringsetu -> fprintf ppf "string.unsafe_set"
  | Pstringrefs -> fprintf ppf "string.get"
  | Pstringsets -> fprintf ppf "string.set"
  | Parraylength _ -> fprintf ppf "array.length"
  | Pmakearray _ -> fprintf ppf "makearray "
  | Parrayrefu _ -> fprintf ppf "array.unsafe_get"
  | Parraysetu _ -> fprintf ppf "array.unsafe_set"
  | Parrayrefs _ -> fprintf ppf "array.get"
  | Parraysets _ -> fprintf ppf "array.set"
  | Pisint -> fprintf ppf "isint"
  | Pisout -> fprintf ppf "isout"
  | Pbittest -> fprintf ppf "testbit"
  | Pbintofint bi -> print_boxed_integer "of_int" ppf bi
  | Pintofbint bi -> print_boxed_integer "to_int" ppf bi
  | Pcvtbint (bi1, bi2) -> print_boxed_integer_conversion ppf bi1 bi2
  | Pnegbint bi -> print_boxed_integer "neg" ppf bi
  | Paddbint bi -> print_boxed_integer "add" ppf bi
  | Psubbint bi -> print_boxed_integer "sub" ppf bi
  | Pmulbint bi -> print_boxed_integer "mul" ppf bi
  | Pdivbint bi -> print_boxed_integer "div" ppf bi
  | Pmodbint bi -> print_boxed_integer "mod" ppf bi
  | Pandbint bi -> print_boxed_integer "and" ppf bi
  | Porbint bi -> print_boxed_integer "or" ppf bi
  | Pxorbint bi -> print_boxed_integer "xor" ppf bi
  | Plslbint bi -> print_boxed_integer "lsl" ppf bi
  | Plsrbint bi -> print_boxed_integer "lsr" ppf bi
  | Pasrbint bi -> print_boxed_integer "asr" ppf bi
  | Pbintcomp(bi, Ceq) -> print_boxed_integer "==" ppf bi
  | Pbintcomp(bi, Cneq) -> print_boxed_integer "!=" ppf bi
  | Pbintcomp(bi, Clt) -> print_boxed_integer "<" ppf bi
  | Pbintcomp(bi, Cgt) -> print_boxed_integer ">" ppf bi
  | Pbintcomp(bi, Cle) -> print_boxed_integer "<=" ppf bi
  | Pbintcomp(bi, Cge) -> print_boxed_integer ">=" ppf bi
  | Pbigarrayref(unsafe, _, kind, layout) -> print_bigarray "get" unsafe kind ppf layout
  | Pbigarrayset(unsafe, _, kind, layout) -> print_bigarray "set" unsafe kind ppf layout

let rec jlambda ppf j =
  let jlams ppf args =
    List.iter (fun l -> fprintf ppf "@ %a" jlambda l) args in
  match j with
  | Jvar id ->
      Ident.print ppf id
  | Jconst cst ->
      Printlambda.structured_constant ppf cst
  | Jdirect_apply ((cls, func), args, _) ->
      fprintf ppf "@[<2>(directapply@ %s@ %s%a)@]" cls func jlams args
  | Jgeneric_apply (func, args, _) ->
      fprintf ppf "@[<2>(genericapply@ %a%a)@]" jlambda func jlams args
  | Jclosure (funcs, vars) ->
      let idents ppf l =
        let spc = ref false in
        List.iter
          (fun id ->
            if !spc then fprintf ppf "@ " else spc := true;
            fprintf ppf "%a" Ident.print id)
          l in
      let fun_decls ppf l =
        List.iter
          (fun ((cls, func), n, ids, j) ->
            fprintf ppf "@ %s@ %s@ %d@ (%a)@ %a"
              cls
              func
              n
              idents ids
              jlambda j)
          l in
      fprintf ppf "@[<2>(closure%a%a)@]"
        fun_decls funcs
        jlams vars
  | Joffset (j, n) ->
      fprintf ppf "@[<2>(offset@ %d@ %a)@]" n jlambda j
  | Jlet (id, expr, body) ->
      let rec letbody = function
        | Jlet (id, arg, body) ->
            fprintf ppf "@ @[<2>%a@ %a@]"
              Ident.print id
              jlambda arg;
            letbody body
        | expr -> expr in
      fprintf ppf "@[<2>(let@ @[<hv 1>(@[<2>%a@ %a@]"
        Ident.print id
        jlambda expr;
      fprintf ppf ")@]@ %a)@]"
        jlambda (letbody body)
  | Jletrec (lst, body) ->
      let bindings ppf l =
        let spc = ref false in
        List.iter
          (fun (id, j) ->
            if !spc then fprintf ppf "@ " else spc := true;
            fprintf ppf "@[<2>%a@ %a@]" Ident.print id jlambda j)
          l in
      fprintf ppf "@[<2>(letrec@ (@[<hv 1>%a@])@ %a)@]"
        bindings lst
        jlambda body
  | Jprim (prim, args, _) ->
      fprintf ppf "@[<2>(prim@ %a%a)@]" primitive prim jlams args
  | Jswitch (arg, sw) ->
      let switch ppf sw =
        let spc = ref false in
        Array.iter
          (fun (n, j) ->
            if !spc then fprintf ppf "@ " else spc := true;
            fprintf ppf "@[<hv 1>case int %i:@ %a@]" n jlambda j)
          (Array.map
             (fun x -> x, sw.js_actions_consts.(x))
             sw.js_index_consts);
        Array.iter
          (fun (n, j) ->
            if !spc then fprintf ppf "@ " else spc := true;
            fprintf ppf "@[<hv 1>case tag %i:@ %a@]" n jlambda j)
          (Array.map
             (fun x -> x, sw.js_actions_blocks.(x))
             sw.js_index_blocks) in
      fprintf ppf "@[<1>(switch %a@ @[<v 0>%a@])@]"
        jlambda arg
        switch sw
  | Jstaticfail (nfail, args) ->
      fprintf ppf "@[<2>(staticfail@ %d%a)@]" nfail jlams args
  | Jstaticcatch (nfail, ids, body, handler) ->
      fprintf ppf "@[(catch@ %a@;<1 -1>with (%d%a)@ %a)@]"
        jlambda body
        nfail
        (fun ppf vars -> match vars with
        | [] -> ()
        | _ -> List.iter (fun x -> fprintf ppf "@ %a" Ident.print x) vars) ids
        jlambda handler
  | Jtrywith (body, id, handler) ->
      fprintf ppf "@[<2>(try@ %a@;<1 -1>with@ %a@ %a)@]"
        jlambda body
        Ident.print id
        jlambda handler
  | Jifthenelse (cond, ifso, ifno) ->
      fprintf ppf "@[<2>(if@ %a@ %a@ %a)@]"
        jlambda cond
        jlambda ifso
        jlambda ifno
  | Jsequence (j1, j2) ->
      fprintf ppf "@[<2>(seq@ %a@ %a)@]"
        jlambda j1
        sequence j2
  | Jwhile (cond, body) ->
      fprintf ppf "@[<2>(while@ %a@ %a)@]"
        jlambda cond
        jlambda body
  | Jfor (id, lo, hi, dir, body) ->
      fprintf ppf "@[<2>(for@ %a@ %a@ %s@ %a@ %a)@]"
        Ident.print id
        jlambda lo
        (if dir = Asttypes.Upto then "to" else "downto")
        jlambda hi
        jlambda body
  | Jassign (id, expr) ->
      fprintf ppf "@[<2>(assign@ %a@ %a)@]" Ident.print id jlambda expr
  | Jsend (kind, meth, obj, args, _) ->
      let kind' = match kind with
      | Lambda.Self -> "self"
      | Lambda.Public -> "public"
      | Lambda.Cached -> "cached" in
      fprintf ppf "@[<2>(send@ %s@ %a@ %a%a)@]"
        kind'
        jlambda obj
        jlambda meth
        jlams args
and sequence ppf = function
  | Jsequence (j1, j2) -> fprintf ppf "%a@ %a" sequence j1 sequence j2
  | j -> jlambda ppf j
