(*
 * This file is part of Cafesterol.
 * Copyright (C) 2007-2009 Xavier Clerc.
 *
 * Cafesterol is free software; you can redistribute it and/or modify
 * it under the terms of the Q Public License as published by
 * Trolltech (with a change to choice of law).
 *
 * Cafesterol is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Q Public License for more details.
 *
 * You should have received a copy of the Q Public License
 * along with this program.  If not, see
 * <http://trolltech.com/products/qt/licenses/licensing/qpl>.
 *)

open BaristaLibrary
open Bytecodeutils


type element =
  | Int
  | Double
  | Block
  | Value
  | Null
  | Runner
  | Exception
  | Undefined

type t = {
    locals : element array;
    operands : element list;
  }

let dummy = { locals = [||]; operands = [] }

let make sz params =
  let len = List.length params in
  { locals = Array.init
      sz
      (fun i ->
        if i = 0 then
          Runner
        else if i < len then
          Value
        else
          Undefined);
    operands = [] }

let translate n = List.map (fun (base, frame) -> (base + n, frame))

let size x = List.length x.operands

let same x = x

let empty x = { locals = x.locals; operands = [] }

let store v i x =
  if (i >= 0) && (i < (Array.length x.locals)) then
    let locals' = Array.copy x.locals in
    locals'.(i) <- v;
    { locals = locals'; operands = x.operands }
  else
    Misc.fatal_error "StackState.store: invalid index"

let store_value = store Value

let store_block = store Block

let store_int = store Int

let store_top i x =
  match x.operands with
  | hd :: _ -> store hd i x
  | _ -> Misc.fatal_error "StackState.store_top: empty stack"

let pop x =
  match x.operands with
  | _ :: tl -> { locals = x.locals; operands = tl }
  | [] -> Misc.fatal_error "StackState.pop: empty stack"

let push v x =
  { locals = x.locals;
    operands = v :: x.operands }

let push_int = push Int

let push_double = push Double

let push_block = push Block

let push_value = push Value

let push_values n x =
  let res = ref x in
  for i = 1 to n do
    res := push_value !res
  done;
  !res

let push_null = push Null

let push_runner = push Runner

let only_exception x = { locals = x.locals; operands = [Exception] }

let push_index i x =
  if (i >= 0) && (i < (Array.length x.locals)) then
    { locals = x.locals; operands = (x.locals.(i)) :: x.operands }
  else
    Misc.fatal_error "StackState.push_index: invalid index"

let push_kind k x =
  match k with
  | Macroinstr.Boxed -> push_value x
  | Macroinstr.Unboxed -> push_int x

let encode_type = function
  | Int -> Attribute.Integer_variable_info
  | Double -> Attribute.Double_variable_info
  | Block -> Attribute.Object_variable_info (`Class_or_interface class_Block)
  | Value -> Attribute.Object_variable_info (`Class_or_interface class_Value)
  | Null -> Attribute.Null_variable_info
  | Runner -> Attribute.Object_variable_info (`Class_or_interface class_AbstractNativeRunner)
  | Exception -> Attribute.Object_variable_info (`Class_or_interface class_FailException)
  | Undefined -> Attribute.Top_variable_info

let generalize old_frame new_frame =
  if old_frame = dummy then
    new_frame
  else if new_frame = dummy then
    old_frame
  else
    let gen = function
      | (Undefined, _)
      | (_, Undefined) -> Undefined
      | (x, y) -> if x = y then x else Undefined in
    let len = Array.length old_frame.locals in
    let locals' = Array.copy old_frame.locals in
    for i = 0 to (pred len) do
      locals'.(i) <- gen (old_frame.locals.(i), new_frame.locals.(i))
    done;
    let operands' = if (List.length old_frame.operands) < (List.length new_frame.operands) then
      new_frame.operands
    else
      old_frame.operands in
    { locals = locals';
      operands = operands' }

let to_stackmap_list init l =
  let encode new_frame delta =
    let locals = List.map encode_type (Array.to_list new_frame.locals) in
    let operands = List.map encode_type new_frame.operands in
    Attribute.Full_frame (delta, locals, List.rev operands) in
  let l' = List.sort compare l in
  let _, l'' = List.fold_left
      (fun (prev, l) elem ->
        let prev_ofs, prev_frame = prev in
        let elem_ofs, elem_frame = elem in
        if prev_ofs = elem_ofs && (l <> []) then
          let gen = generalize prev_frame elem_frame in
          ((prev_ofs, gen), (prev_ofs, gen) :: (List.tl l))
        else
          (elem, elem :: l))
      ((0, init), [])
      l' in
  List.rev_map (fun (ofs, elem) -> encode elem ofs) l''
