(*
 * This file is part of Cafesterol.
 * Copyright (C) 2007-2009 Xavier Clerc.
 *
 * Cafesterol is free software; you can redistribute it and/or modify
 * it under the terms of the Q Public License as published by
 * Trolltech (with a change to choice of law).
 *
 * Cafesterol is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Q Public License for more details.
 *
 * You should have received a copy of the Q Public License
 * along with this program.  If not, see
 * <http://trolltech.com/products/qt/licenses/licensing/qpl>.
 *)

(* Stack states used to generate Java stack maps. *)

type t
(** State of the stack (both locals and operand stack) at a given point in code. *)

val dummy : t
(** The stack state with no local, and an empty operand stack .*)

val make : int -> BaristaLibrary.Descriptor.for_parameter list -> t
(** Builds a new state from a list of parameters defining the signature of a static method.
    The passed [int] gives the size of locals and should hence be greater than or equal to
    the size of the passed list. *)

val translate : int -> ((int * t) list) -> ((int * t) list)
(** [translate n l] returns a modification of [l] where each [(o, x)] couple is transformed
    into [(o + n, x)]. *)

val size : t -> int
(** Returns the size of the operand stack. *)

val same : t -> t
(** Identity function. *)

val empty : t -> t
(** Returns a copy of the passed stack state where the operand stack is empty. *)

val store_value : int -> t -> t
(** Returns a copy of the passed stack state where a {i Value} is stored at the passed index of the locals. *)

val store_block : int -> t -> t
(** Returns a copy of the passed stack state where a {i Block} is stored at the passed index of the locals. *)

val store_int : int -> t -> t
(** Returns a copy of the passed stack state where an {i int} is stored at the passed index of the locals. *)

val store_top : int -> t -> t
(** Returns a copy of the passed stack state where an element having the type of the operand stack top is stored at the passed index of the locals. *)

val pop : t -> t
(** Returns a copy of the passed stack state where the top element of the operand stack has been removed. *)

val push_int : t -> t
(** Returns a copy of the passed stack state where an {i int} is added onto the top of the operand stack. *)

val push_double : t -> t
(** Returns a copy of the passed stack state where a {i double} is added onto the top of the operand stack. *)

val push_block : t -> t
(** Returns a copy of the passed stack state where a {i block} is added onto the top of the operand stack. *)

val push_value : t -> t
(** Returns a copy of the passed stack state where a {i Value} is added onto the top of the operand stack. *)

val push_values : int -> t -> t
(** Returns a copy of the passed stack state where [n] {i Value}s are added onto the top of the operand stack. *)

val push_null : t -> t
(** Returns a copy of the passed stack state where a {i null} is added onto the top of the operand stack. *)

val push_runner : t -> t
(** Returns a copy of the passed stack state where an {i AbstractNativeCodeRunner} is added onto the top of the operand stack. *)

val only_exception : t -> t
(** Returns a copy of the passed stack state where a the operand stack is reduced to a single {i Fail.Exception} element. *)

val push_kind : Macroinstr.boxed -> t -> t
(** Returns a copy of the passed stack state where either an {i int} or a {i Value} is added onto the top of the operand stack. *)

val push_index : int -> t -> t
(** Returns a copy of the passed stack state where a {i T} is added onto the top of the operand stack.,
    {i T} being the type of the local at the passed index. *)

val generalize : t -> t -> t
(** [generalize x y] returns the generalization of [x] and [y].
    The generalization has the longuest operand stack and locals compatible with
    both stack states. *)

val to_stackmap_list : t -> (int * t) list -> BaristaLibrary.Attribute.stack_map_frame list
(** [to_stackmap_list i l] converts the list [l] of [(offset, state)] couples into
    a Java stack map information. [i] is the initial stack state of the method. *)
